# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: oef_search.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="oef_search.proto",
    package="fetch.aea.OefSearch",
    syntax="proto3",
    serialized_options=None,
    serialized_pb=b'\n\x10oef_search.proto\x12\x13\x66\x65tch.aea.OefSearch"\xdf\r\n\x10OefSearchMessage\x12\x12\n\nmessage_id\x18\x01 \x01(\x05\x12"\n\x1a\x64ialogue_starter_reference\x18\x02 \x01(\t\x12$\n\x1c\x64ialogue_responder_reference\x18\x03 \x01(\t\x12\x0e\n\x06target\x18\x04 \x01(\x05\x12Q\n\toef_error\x18\x05 \x01(\x0b\x32<.fetch.aea.OefSearch.OefSearchMessage.Oef_Error_PerformativeH\x00\x12_\n\x10register_service\x18\x06 \x01(\x0b\x32\x43.fetch.aea.OefSearch.OefSearchMessage.Register_Service_PerformativeH\x00\x12Y\n\rsearch_result\x18\x07 \x01(\x0b\x32@.fetch.aea.OefSearch.OefSearchMessage.Search_Result_PerformativeH\x00\x12]\n\x0fsearch_services\x18\x08 \x01(\x0b\x32\x42.fetch.aea.OefSearch.OefSearchMessage.Search_Services_PerformativeH\x00\x12M\n\x07success\x18\t \x01(\x0b\x32:.fetch.aea.OefSearch.OefSearchMessage.Success_PerformativeH\x00\x12\x63\n\x12unregister_service\x18\n \x01(\x0b\x32\x45.fetch.aea.OefSearch.OefSearchMessage.Unregister_Service_PerformativeH\x00\x1a!\n\nAgentsInfo\x12\x13\n\x0b\x61gents_info\x18\x01 \x01(\x0c\x1a"\n\x0b\x44\x65scription\x12\x13\n\x0b\x64\x65scription\x18\x01 \x01(\x0c\x1a\xd1\x01\n\x11OefErrorOperation\x12W\n\toef_error\x18\x01 \x01(\x0e\x32\x44.fetch.aea.OefSearch.OefSearchMessage.OefErrorOperation.OefErrorEnum"c\n\x0cOefErrorEnum\x12\x14\n\x10REGISTER_SERVICE\x10\x00\x12\x16\n\x12UNREGISTER_SERVICE\x10\x01\x12\x13\n\x0fSEARCH_SERVICES\x10\x02\x12\x10\n\x0cSEND_MESSAGE\x10\x03\x1a\x8b\x01\n\x05Query\x12\x0f\n\x05\x62ytes\x18\x01 \x01(\x0cH\x00\x12\x46\n\x07nothing\x18\x02 \x01(\x0b\x32\x33.fetch.aea.OefSearch.OefSearchMessage.Query.NothingH\x00\x12\x15\n\x0bquery_bytes\x18\x03 \x01(\x0cH\x00\x1a\t\n\x07NothingB\x07\n\x05query\x1ao\n\x1dRegister_Service_Performative\x12N\n\x13service_description\x18\x01 \x01(\x0b\x32\x31.fetch.aea.OefSearch.OefSearchMessage.Description\x1aq\n\x1fUnregister_Service_Performative\x12N\n\x13service_description\x18\x01 \x01(\x0b\x32\x31.fetch.aea.OefSearch.OefSearchMessage.Description\x1aZ\n\x1cSearch_Services_Performative\x12:\n\x05query\x18\x01 \x01(\x0b\x32+.fetch.aea.OefSearch.OefSearchMessage.Query\x1as\n\x1aSearch_Result_Performative\x12\x0e\n\x06\x61gents\x18\x01 \x03(\t\x12\x45\n\x0b\x61gents_info\x18\x02 \x01(\x0b\x32\x30.fetch.aea.OefSearch.OefSearchMessage.AgentsInfo\x1a]\n\x14Success_Performative\x12\x45\n\x0b\x61gents_info\x18\x01 \x01(\x0b\x32\x30.fetch.aea.OefSearch.OefSearchMessage.AgentsInfo\x1an\n\x16Oef_Error_Performative\x12T\n\x13oef_error_operation\x18\x01 \x01(\x0b\x32\x37.fetch.aea.OefSearch.OefSearchMessage.OefErrorOperationB\x0e\n\x0cperformativeb\x06proto3',
)


_OEFSEARCHMESSAGE_OEFERROROPERATION_OEFERRORENUM = _descriptor.EnumDescriptor(
    name="OefErrorEnum",
    full_name="fetch.aea.OefSearch.OefSearchMessage.OefErrorOperation.OefErrorEnum",
    filename=None,
    file=DESCRIPTOR,
    values=[
        _descriptor.EnumValueDescriptor(
            name="REGISTER_SERVICE",
            index=0,
            number=0,
            serialized_options=None,
            type=None,
        ),
        _descriptor.EnumValueDescriptor(
            name="UNREGISTER_SERVICE",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
        ),
        _descriptor.EnumValueDescriptor(
            name="SEARCH_SERVICES",
            index=2,
            number=2,
            serialized_options=None,
            type=None,
        ),
        _descriptor.EnumValueDescriptor(
            name="SEND_MESSAGE", index=3, number=3, serialized_options=None, type=None
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=900,
    serialized_end=999,
)
_sym_db.RegisterEnumDescriptor(_OEFSEARCHMESSAGE_OEFERROROPERATION_OEFERRORENUM)


_OEFSEARCHMESSAGE_AGENTSINFO = _descriptor.Descriptor(
    name="AgentsInfo",
    full_name="fetch.aea.OefSearch.OefSearchMessage.AgentsInfo",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="agents_info",
            full_name="fetch.aea.OefSearch.OefSearchMessage.AgentsInfo.agents_info",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=718,
    serialized_end=751,
)

_OEFSEARCHMESSAGE_DESCRIPTION = _descriptor.Descriptor(
    name="Description",
    full_name="fetch.aea.OefSearch.OefSearchMessage.Description",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="description",
            full_name="fetch.aea.OefSearch.OefSearchMessage.Description.description",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=753,
    serialized_end=787,
)

_OEFSEARCHMESSAGE_OEFERROROPERATION = _descriptor.Descriptor(
    name="OefErrorOperation",
    full_name="fetch.aea.OefSearch.OefSearchMessage.OefErrorOperation",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="oef_error",
            full_name="fetch.aea.OefSearch.OefSearchMessage.OefErrorOperation.oef_error",
            index=0,
            number=1,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[_OEFSEARCHMESSAGE_OEFERROROPERATION_OEFERRORENUM,],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=790,
    serialized_end=999,
)

_OEFSEARCHMESSAGE_QUERY_NOTHING = _descriptor.Descriptor(
    name="Nothing",
    full_name="fetch.aea.OefSearch.OefSearchMessage.Query.Nothing",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1123,
    serialized_end=1132,
)

_OEFSEARCHMESSAGE_QUERY = _descriptor.Descriptor(
    name="Query",
    full_name="fetch.aea.OefSearch.OefSearchMessage.Query",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="bytes",
            full_name="fetch.aea.OefSearch.OefSearchMessage.Query.bytes",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="nothing",
            full_name="fetch.aea.OefSearch.OefSearchMessage.Query.nothing",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="query_bytes",
            full_name="fetch.aea.OefSearch.OefSearchMessage.Query.query_bytes",
            index=2,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[_OEFSEARCHMESSAGE_QUERY_NOTHING,],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="query",
            full_name="fetch.aea.OefSearch.OefSearchMessage.Query.query",
            index=0,
            containing_type=None,
            fields=[],
        ),
    ],
    serialized_start=1002,
    serialized_end=1141,
)

_OEFSEARCHMESSAGE_REGISTER_SERVICE_PERFORMATIVE = _descriptor.Descriptor(
    name="Register_Service_Performative",
    full_name="fetch.aea.OefSearch.OefSearchMessage.Register_Service_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="service_description",
            full_name="fetch.aea.OefSearch.OefSearchMessage.Register_Service_Performative.service_description",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1143,
    serialized_end=1254,
)

_OEFSEARCHMESSAGE_UNREGISTER_SERVICE_PERFORMATIVE = _descriptor.Descriptor(
    name="Unregister_Service_Performative",
    full_name="fetch.aea.OefSearch.OefSearchMessage.Unregister_Service_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="service_description",
            full_name="fetch.aea.OefSearch.OefSearchMessage.Unregister_Service_Performative.service_description",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1256,
    serialized_end=1369,
)

_OEFSEARCHMESSAGE_SEARCH_SERVICES_PERFORMATIVE = _descriptor.Descriptor(
    name="Search_Services_Performative",
    full_name="fetch.aea.OefSearch.OefSearchMessage.Search_Services_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="query",
            full_name="fetch.aea.OefSearch.OefSearchMessage.Search_Services_Performative.query",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1371,
    serialized_end=1461,
)

_OEFSEARCHMESSAGE_SEARCH_RESULT_PERFORMATIVE = _descriptor.Descriptor(
    name="Search_Result_Performative",
    full_name="fetch.aea.OefSearch.OefSearchMessage.Search_Result_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="agents",
            full_name="fetch.aea.OefSearch.OefSearchMessage.Search_Result_Performative.agents",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="agents_info",
            full_name="fetch.aea.OefSearch.OefSearchMessage.Search_Result_Performative.agents_info",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1463,
    serialized_end=1578,
)

_OEFSEARCHMESSAGE_SUCCESS_PERFORMATIVE = _descriptor.Descriptor(
    name="Success_Performative",
    full_name="fetch.aea.OefSearch.OefSearchMessage.Success_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="agents_info",
            full_name="fetch.aea.OefSearch.OefSearchMessage.Success_Performative.agents_info",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1580,
    serialized_end=1673,
)

_OEFSEARCHMESSAGE_OEF_ERROR_PERFORMATIVE = _descriptor.Descriptor(
    name="Oef_Error_Performative",
    full_name="fetch.aea.OefSearch.OefSearchMessage.Oef_Error_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="oef_error_operation",
            full_name="fetch.aea.OefSearch.OefSearchMessage.Oef_Error_Performative.oef_error_operation",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1675,
    serialized_end=1785,
)

_OEFSEARCHMESSAGE = _descriptor.Descriptor(
    name="OefSearchMessage",
    full_name="fetch.aea.OefSearch.OefSearchMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="message_id",
            full_name="fetch.aea.OefSearch.OefSearchMessage.message_id",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dialogue_starter_reference",
            full_name="fetch.aea.OefSearch.OefSearchMessage.dialogue_starter_reference",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dialogue_responder_reference",
            full_name="fetch.aea.OefSearch.OefSearchMessage.dialogue_responder_reference",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="target",
            full_name="fetch.aea.OefSearch.OefSearchMessage.target",
            index=3,
            number=4,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="oef_error",
            full_name="fetch.aea.OefSearch.OefSearchMessage.oef_error",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="register_service",
            full_name="fetch.aea.OefSearch.OefSearchMessage.register_service",
            index=5,
            number=6,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="search_result",
            full_name="fetch.aea.OefSearch.OefSearchMessage.search_result",
            index=6,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="search_services",
            full_name="fetch.aea.OefSearch.OefSearchMessage.search_services",
            index=7,
            number=8,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="success",
            full_name="fetch.aea.OefSearch.OefSearchMessage.success",
            index=8,
            number=9,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="unregister_service",
            full_name="fetch.aea.OefSearch.OefSearchMessage.unregister_service",
            index=9,
            number=10,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[
        _OEFSEARCHMESSAGE_AGENTSINFO,
        _OEFSEARCHMESSAGE_DESCRIPTION,
        _OEFSEARCHMESSAGE_OEFERROROPERATION,
        _OEFSEARCHMESSAGE_QUERY,
        _OEFSEARCHMESSAGE_REGISTER_SERVICE_PERFORMATIVE,
        _OEFSEARCHMESSAGE_UNREGISTER_SERVICE_PERFORMATIVE,
        _OEFSEARCHMESSAGE_SEARCH_SERVICES_PERFORMATIVE,
        _OEFSEARCHMESSAGE_SEARCH_RESULT_PERFORMATIVE,
        _OEFSEARCHMESSAGE_SUCCESS_PERFORMATIVE,
        _OEFSEARCHMESSAGE_OEF_ERROR_PERFORMATIVE,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="performative",
            full_name="fetch.aea.OefSearch.OefSearchMessage.performative",
            index=0,
            containing_type=None,
            fields=[],
        ),
    ],
    serialized_start=42,
    serialized_end=1801,
)

_OEFSEARCHMESSAGE_AGENTSINFO.containing_type = _OEFSEARCHMESSAGE
_OEFSEARCHMESSAGE_DESCRIPTION.containing_type = _OEFSEARCHMESSAGE
_OEFSEARCHMESSAGE_OEFERROROPERATION.fields_by_name[
    "oef_error"
].enum_type = _OEFSEARCHMESSAGE_OEFERROROPERATION_OEFERRORENUM
_OEFSEARCHMESSAGE_OEFERROROPERATION.containing_type = _OEFSEARCHMESSAGE
_OEFSEARCHMESSAGE_OEFERROROPERATION_OEFERRORENUM.containing_type = (
    _OEFSEARCHMESSAGE_OEFERROROPERATION
)
_OEFSEARCHMESSAGE_QUERY_NOTHING.containing_type = _OEFSEARCHMESSAGE_QUERY
_OEFSEARCHMESSAGE_QUERY.fields_by_name[
    "nothing"
].message_type = _OEFSEARCHMESSAGE_QUERY_NOTHING
_OEFSEARCHMESSAGE_QUERY.containing_type = _OEFSEARCHMESSAGE
_OEFSEARCHMESSAGE_QUERY.oneofs_by_name["query"].fields.append(
    _OEFSEARCHMESSAGE_QUERY.fields_by_name["bytes"]
)
_OEFSEARCHMESSAGE_QUERY.fields_by_name[
    "bytes"
].containing_oneof = _OEFSEARCHMESSAGE_QUERY.oneofs_by_name["query"]
_OEFSEARCHMESSAGE_QUERY.oneofs_by_name["query"].fields.append(
    _OEFSEARCHMESSAGE_QUERY.fields_by_name["nothing"]
)
_OEFSEARCHMESSAGE_QUERY.fields_by_name[
    "nothing"
].containing_oneof = _OEFSEARCHMESSAGE_QUERY.oneofs_by_name["query"]
_OEFSEARCHMESSAGE_QUERY.oneofs_by_name["query"].fields.append(
    _OEFSEARCHMESSAGE_QUERY.fields_by_name["query_bytes"]
)
_OEFSEARCHMESSAGE_QUERY.fields_by_name[
    "query_bytes"
].containing_oneof = _OEFSEARCHMESSAGE_QUERY.oneofs_by_name["query"]
_OEFSEARCHMESSAGE_REGISTER_SERVICE_PERFORMATIVE.fields_by_name[
    "service_description"
].message_type = _OEFSEARCHMESSAGE_DESCRIPTION
_OEFSEARCHMESSAGE_REGISTER_SERVICE_PERFORMATIVE.containing_type = _OEFSEARCHMESSAGE
_OEFSEARCHMESSAGE_UNREGISTER_SERVICE_PERFORMATIVE.fields_by_name[
    "service_description"
].message_type = _OEFSEARCHMESSAGE_DESCRIPTION
_OEFSEARCHMESSAGE_UNREGISTER_SERVICE_PERFORMATIVE.containing_type = _OEFSEARCHMESSAGE
_OEFSEARCHMESSAGE_SEARCH_SERVICES_PERFORMATIVE.fields_by_name[
    "query"
].message_type = _OEFSEARCHMESSAGE_QUERY
_OEFSEARCHMESSAGE_SEARCH_SERVICES_PERFORMATIVE.containing_type = _OEFSEARCHMESSAGE
_OEFSEARCHMESSAGE_SEARCH_RESULT_PERFORMATIVE.fields_by_name[
    "agents_info"
].message_type = _OEFSEARCHMESSAGE_AGENTSINFO
_OEFSEARCHMESSAGE_SEARCH_RESULT_PERFORMATIVE.containing_type = _OEFSEARCHMESSAGE
_OEFSEARCHMESSAGE_SUCCESS_PERFORMATIVE.fields_by_name[
    "agents_info"
].message_type = _OEFSEARCHMESSAGE_AGENTSINFO
_OEFSEARCHMESSAGE_SUCCESS_PERFORMATIVE.containing_type = _OEFSEARCHMESSAGE
_OEFSEARCHMESSAGE_OEF_ERROR_PERFORMATIVE.fields_by_name[
    "oef_error_operation"
].message_type = _OEFSEARCHMESSAGE_OEFERROROPERATION
_OEFSEARCHMESSAGE_OEF_ERROR_PERFORMATIVE.containing_type = _OEFSEARCHMESSAGE
_OEFSEARCHMESSAGE.fields_by_name[
    "oef_error"
].message_type = _OEFSEARCHMESSAGE_OEF_ERROR_PERFORMATIVE
_OEFSEARCHMESSAGE.fields_by_name[
    "register_service"
].message_type = _OEFSEARCHMESSAGE_REGISTER_SERVICE_PERFORMATIVE
_OEFSEARCHMESSAGE.fields_by_name[
    "search_result"
].message_type = _OEFSEARCHMESSAGE_SEARCH_RESULT_PERFORMATIVE
_OEFSEARCHMESSAGE.fields_by_name[
    "search_services"
].message_type = _OEFSEARCHMESSAGE_SEARCH_SERVICES_PERFORMATIVE
_OEFSEARCHMESSAGE.fields_by_name[
    "success"
].message_type = _OEFSEARCHMESSAGE_SUCCESS_PERFORMATIVE
_OEFSEARCHMESSAGE.fields_by_name[
    "unregister_service"
].message_type = _OEFSEARCHMESSAGE_UNREGISTER_SERVICE_PERFORMATIVE
_OEFSEARCHMESSAGE.oneofs_by_name["performative"].fields.append(
    _OEFSEARCHMESSAGE.fields_by_name["oef_error"]
)
_OEFSEARCHMESSAGE.fields_by_name[
    "oef_error"
].containing_oneof = _OEFSEARCHMESSAGE.oneofs_by_name["performative"]
_OEFSEARCHMESSAGE.oneofs_by_name["performative"].fields.append(
    _OEFSEARCHMESSAGE.fields_by_name["register_service"]
)
_OEFSEARCHMESSAGE.fields_by_name[
    "register_service"
].containing_oneof = _OEFSEARCHMESSAGE.oneofs_by_name["performative"]
_OEFSEARCHMESSAGE.oneofs_by_name["performative"].fields.append(
    _OEFSEARCHMESSAGE.fields_by_name["search_result"]
)
_OEFSEARCHMESSAGE.fields_by_name[
    "search_result"
].containing_oneof = _OEFSEARCHMESSAGE.oneofs_by_name["performative"]
_OEFSEARCHMESSAGE.oneofs_by_name["performative"].fields.append(
    _OEFSEARCHMESSAGE.fields_by_name["search_services"]
)
_OEFSEARCHMESSAGE.fields_by_name[
    "search_services"
].containing_oneof = _OEFSEARCHMESSAGE.oneofs_by_name["performative"]
_OEFSEARCHMESSAGE.oneofs_by_name["performative"].fields.append(
    _OEFSEARCHMESSAGE.fields_by_name["success"]
)
_OEFSEARCHMESSAGE.fields_by_name[
    "success"
].containing_oneof = _OEFSEARCHMESSAGE.oneofs_by_name["performative"]
_OEFSEARCHMESSAGE.oneofs_by_name["performative"].fields.append(
    _OEFSEARCHMESSAGE.fields_by_name["unregister_service"]
)
_OEFSEARCHMESSAGE.fields_by_name[
    "unregister_service"
].containing_oneof = _OEFSEARCHMESSAGE.oneofs_by_name["performative"]
DESCRIPTOR.message_types_by_name["OefSearchMessage"] = _OEFSEARCHMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

OefSearchMessage = _reflection.GeneratedProtocolMessageType(
    "OefSearchMessage",
    (_message.Message,),
    {
        "AgentsInfo": _reflection.GeneratedProtocolMessageType(
            "AgentsInfo",
            (_message.Message,),
            {
                "DESCRIPTOR": _OEFSEARCHMESSAGE_AGENTSINFO,
                "__module__": "oef_search_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.OefSearch.OefSearchMessage.AgentsInfo)
            },
        ),
        "Description": _reflection.GeneratedProtocolMessageType(
            "Description",
            (_message.Message,),
            {
                "DESCRIPTOR": _OEFSEARCHMESSAGE_DESCRIPTION,
                "__module__": "oef_search_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.OefSearch.OefSearchMessage.Description)
            },
        ),
        "OefErrorOperation": _reflection.GeneratedProtocolMessageType(
            "OefErrorOperation",
            (_message.Message,),
            {
                "DESCRIPTOR": _OEFSEARCHMESSAGE_OEFERROROPERATION,
                "__module__": "oef_search_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.OefSearch.OefSearchMessage.OefErrorOperation)
            },
        ),
        "Query": _reflection.GeneratedProtocolMessageType(
            "Query",
            (_message.Message,),
            {
                "Nothing": _reflection.GeneratedProtocolMessageType(
                    "Nothing",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _OEFSEARCHMESSAGE_QUERY_NOTHING,
                        "__module__": "oef_search_pb2"
                        # @@protoc_insertion_point(class_scope:fetch.aea.OefSearch.OefSearchMessage.Query.Nothing)
                    },
                ),
                "DESCRIPTOR": _OEFSEARCHMESSAGE_QUERY,
                "__module__": "oef_search_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.OefSearch.OefSearchMessage.Query)
            },
        ),
        "Register_Service_Performative": _reflection.GeneratedProtocolMessageType(
            "Register_Service_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _OEFSEARCHMESSAGE_REGISTER_SERVICE_PERFORMATIVE,
                "__module__": "oef_search_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.OefSearch.OefSearchMessage.Register_Service_Performative)
            },
        ),
        "Unregister_Service_Performative": _reflection.GeneratedProtocolMessageType(
            "Unregister_Service_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _OEFSEARCHMESSAGE_UNREGISTER_SERVICE_PERFORMATIVE,
                "__module__": "oef_search_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.OefSearch.OefSearchMessage.Unregister_Service_Performative)
            },
        ),
        "Search_Services_Performative": _reflection.GeneratedProtocolMessageType(
            "Search_Services_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _OEFSEARCHMESSAGE_SEARCH_SERVICES_PERFORMATIVE,
                "__module__": "oef_search_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.OefSearch.OefSearchMessage.Search_Services_Performative)
            },
        ),
        "Search_Result_Performative": _reflection.GeneratedProtocolMessageType(
            "Search_Result_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _OEFSEARCHMESSAGE_SEARCH_RESULT_PERFORMATIVE,
                "__module__": "oef_search_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.OefSearch.OefSearchMessage.Search_Result_Performative)
            },
        ),
        "Success_Performative": _reflection.GeneratedProtocolMessageType(
            "Success_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _OEFSEARCHMESSAGE_SUCCESS_PERFORMATIVE,
                "__module__": "oef_search_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.OefSearch.OefSearchMessage.Success_Performative)
            },
        ),
        "Oef_Error_Performative": _reflection.GeneratedProtocolMessageType(
            "Oef_Error_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _OEFSEARCHMESSAGE_OEF_ERROR_PERFORMATIVE,
                "__module__": "oef_search_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.OefSearch.OefSearchMessage.Oef_Error_Performative)
            },
        ),
        "DESCRIPTOR": _OEFSEARCHMESSAGE,
        "__module__": "oef_search_pb2"
        # @@protoc_insertion_point(class_scope:fetch.aea.OefSearch.OefSearchMessage)
    },
)
_sym_db.RegisterMessage(OefSearchMessage)
_sym_db.RegisterMessage(OefSearchMessage.AgentsInfo)
_sym_db.RegisterMessage(OefSearchMessage.Description)
_sym_db.RegisterMessage(OefSearchMessage.OefErrorOperation)
_sym_db.RegisterMessage(OefSearchMessage.Query)
_sym_db.RegisterMessage(OefSearchMessage.Query.Nothing)
_sym_db.RegisterMessage(OefSearchMessage.Register_Service_Performative)
_sym_db.RegisterMessage(OefSearchMessage.Unregister_Service_Performative)
_sym_db.RegisterMessage(OefSearchMessage.Search_Services_Performative)
_sym_db.RegisterMessage(OefSearchMessage.Search_Result_Performative)
_sym_db.RegisterMessage(OefSearchMessage.Success_Performative)
_sym_db.RegisterMessage(OefSearchMessage.Oef_Error_Performative)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Key pieces of functionality for CLI GUI."""

import glob
import os
import subprocess  # nosec
import sys
import threading
from typing import Any, Dict, List, Tuple

import connexion
import flask
from click import ClickException

from aea.cli.add import add_item as cli_add_item
from aea.cli.create import create_aea as cli_create_aea
from aea.cli.delete import delete_aea as cli_delete_aea
from aea.cli.fetch import fetch_agent_locally as cli_fetch_agent_locally
from aea.cli.list import list_agent_items as cli_list_agent_items
from aea.cli.registry.fetch import fetch_agent as cli_fetch_agent
from aea.cli.remove import remove_item as cli_remove_item
from aea.cli.scaffold import scaffold_item as cli_scaffold_item
from aea.cli.search import search_items as cli_search_items
from aea.cli.search import setup_search_ctx as cli_setup_search_ctx
from aea.cli.utils.config import try_to_load_agent_config
from aea.cli.utils.context import Context
from aea.cli.utils.formatting import sort_items
from aea.cli_gui.utils import (
    ProcessState,
    call_aea_async,
    get_process_status,
    is_agent_dir,
    read_error,
    read_tty,
    stop_agent_process,
    terminate_processes,
)
from aea.configurations.base import PublicId


elements = [
    ["local", "agent", "localAgents"],
    ["registered", "protocol", "registeredProtocols"],
    ["registered", "connection", "registeredConections"],
    ["registered", "skill", "registeredSkills"],
    ["local", "protocol", "localProtocols"],
    ["local", "connection", "localConnections"],
    ["local", "contract", "localContracts"],
    ["local", "skill", "localSkills"],
]


max_log_lines = 100


class AppContext:  # pylint: disable=too-few-public-methods
    """Store useful global information about the app.

    Can't add it into the app object itself because mypy complains.
    """

    agent_processes: Dict[str, subprocess.Popen] = {}
    agent_tty: Dict[str, List[str]] = {}
    agent_error: Dict[str, List[str]] = {}

    ui_is_starting = False
    agents_dir = os.path.abspath(os.getcwd())
    module_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "../../")

    local = "--local" in sys.argv  # a hack to get "local" option from cli args


app_context = AppContext()


def get_agents() -> List[Dict]:
    """Return list of all local agents."""
    file_list = glob.glob(os.path.join(app_context.agents_dir, "*"))

    agent_list = []

    for path in file_list:
        if is_agent_dir(path):
            _head, tail = os.path.split(path)
            agent_list.append(
                {
                    "public_id": tail,  # it is not a public_id actually, just a folder name.
                    # the reason it's called here so is the view that is used to represent items with public_ids
                    # used also for agent displaying
                    # change it when we will have a separate view for an agent.
                    "description": "placeholder description",
                }
            )

    return agent_list


def get_registered_items(item_type: str):
    """Create a new AEA project."""
    # need to place ourselves one directory down so the cher can find the packages
    ctx = Context(cwd=app_context.agents_dir)
    try:
        cli_setup_search_ctx(ctx, local=app_context.local)
        result, _ = cli_search_items(ctx, item_type, query="", page=1)
    except ClickException:
        return {"detail": "Failed to search items."}, 400  # 400 Bad request
    else:
        sorted_items = sort_items(result)
        return sorted_items, 200  # 200 (Success)


def search_registered_items(item_type: str, search_term: str):
    """Create a new AEA project."""
    # need to place ourselves one directory down so the searcher can find the packages
    ctx = Context(cwd=app_context.agents_dir)
    try:
        cli_setup_search_ctx(ctx, local=app_context.local)
        result, _ = cli_search_items(ctx, item_type, query=search_term, page=1)
    except ClickException:
        return {"detail": "Failed to search items."}, 400  # 400 Bad request
    else:
        sorted_items = sort_items(result)
        response = {
            "search_result": sorted_items,
            "item_type": item_type,
            "search_term": search_term,
        }
        return response, 200  # 200 (Success)


def create_agent(agent_id: str):
    """Create a new AEA project."""
    ctx = Context(cwd=app_context.agents_dir)
    try:
        cli_create_aea(ctx, agent_id, local=app_context.local)
    except ClickException as e:
        return (
            {"detail": "Failed to create Agent. {}".format(str(e))},
            400,
        )  # 400 Bad request
    else:
        return agent_id, 201  # 201 (Created)


def delete_agent(agent_id: str):
    """Delete an existing AEA project."""
    ctx = Context(cwd=app_context.agents_dir)
    try:
        cli_delete_aea(ctx, agent_id)
    except ClickException:
        return (
            {"detail": "Failed to delete Agent {} - it may not exist".format(agent_id)},
            400,
        )  # 400 Bad request
    else:
        return "Agent {} deleted".format(agent_id), 200  # 200 (OK)


def add_item(agent_id: str, item_type: str, item_id: str):
    """Add a protocol, skill or connection to the register to a local agent."""
    ctx = Context(cwd=os.path.join(app_context.agents_dir, agent_id))
    ctx.set_config("is_local", app_context.local)
    try:
        try_to_load_agent_config(ctx)
        cli_add_item(ctx, item_type, PublicId.from_str(item_id))
    except ClickException as e:
        return (
            {
                "detail": "Failed to add {} {} to agent {}. {}".format(
                    item_type, item_id, agent_id, str(e)
                )
            },
            400,
        )  # 400 Bad request
    else:
        return agent_id, 201  # 200 (OK)


def fetch_agent(agent_id: str):
    """Fetch an agent."""
    ctx = Context(cwd=app_context.agents_dir)
    fetch_agent_ = cli_fetch_agent_locally if app_context.local else cli_fetch_agent
    try:
        agent_public_id = PublicId.from_str(agent_id)
        fetch_agent_(ctx, agent_public_id)
    except ClickException as e:
        return (
            {"detail": "Failed to fetch an agent {}. {}".format(agent_id, str(e))},
            400,
        )  # 400 Bad request
    else:
        return agent_public_id.name, 201  # 200 (OK)


def remove_local_item(agent_id: str, item_type: str, item_id: str):
    """Remove a protocol, skill or connection from a local agent."""
    agent_dir = os.path.join(app_context.agents_dir, agent_id)
    ctx = Context(cwd=agent_dir)
    try:
        try_to_load_agent_config(ctx)
        cli_remove_item(ctx, item_type, PublicId.from_str(item_id))
    except ClickException:
        return (
            {
                "detail": "Failed to remove {} {} from agent {}".format(
                    item_type, item_id, agent_id
                )
            },
            400,
        )  # 400 Bad request
    else:
        return agent_id, 201  # 200 (OK)


def get_local_items(agent_id: str, item_type: str) -> Tuple[List[Dict[Any, Any]], int]:
    """
    Return a list of protocols, skills or connections supported by a local agent.

    :param agent_id: the id of the agent
    :param item_type: the type of item
    """
    if agent_id == "NONE":
        return [], 200  # 200 (Success)

    # need to place ourselves one directory down so the searcher can find the packages
    ctx = Context(cwd=os.path.join(app_context.agents_dir, agent_id))
    try:
        try_to_load_agent_config(ctx)
        result = cli_list_agent_items(ctx, item_type)
    except ClickException:
        return [{"detail": "Failed to list agent items."}], 400  # 400 Bad request
    else:
        sorted_items = sort_items(result)
        return sorted_items, 200  # 200 (Success)


def scaffold_item(agent_id: str, item_type: str, item_id: str):
    """Scaffold a moslty empty item on an agent (either protocol, skill or connection)."""
    agent_dir = os.path.join(app_context.agents_dir, agent_id)
    ctx = Context(cwd=agent_dir)
    try:
        try_to_load_agent_config(ctx)
        cli_scaffold_item(ctx, item_type, item_id)
    except ClickException:
        return (
            {
                "detail": "Failed to scaffold a new {} in to agent {}".format(
                    item_type, agent_id
                )
            },
            400,
        )  # 400 Bad request
    else:
        return agent_id, 201  # 200 (OK)


def start_agent(agent_id: str, connection_id: PublicId):
    """Start a local agent running."""
    # Test if it is already running in some form
    if agent_id in app_context.agent_processes:
        if (
            get_process_status(app_context.agent_processes[agent_id])
            != ProcessState.RUNNING
        ):  # pragma: no cover
            if app_context.agent_processes[agent_id] is not None:
                app_context.agent_processes[agent_id].terminate()
                app_context.agent_processes[agent_id].wait()
            del app_context.agent_processes[agent_id]
            del app_context.agent_tty[agent_id]
            del app_context.agent_error[agent_id]
        else:
            return (
                {"detail": "Agent {} is already running".format(agent_id)},
                400,
            )  # 400 Bad request

    agent_dir = os.path.join(app_context.agents_dir, agent_id)

    if connection_id is not None and connection_id != "":
        connections = get_local_items(agent_id, "connection")[0]
        has_named_connection = False
        for element in connections:
            if element["public_id"] == connection_id:
                has_named_connection = True
        if has_named_connection:
            agent_process = call_aea_async(
                [
                    sys.executable,
                    "-m",
                    "aea.cli",
                    "run",
                    "--connections",
                    str(connection_id),
                ],
                agent_dir,
            )
        else:
            return (
                {
                    "detail": "Trying to run agent {} with non-existent connection: {}".format(
                        agent_id, connection_id
                    )
                },
                400,
            )  # 400 Bad request
    else:
        agent_process = call_aea_async(
            [sys.executable, "-m", "aea.cli", "run", "--install-deps"], agent_dir
        )

    if agent_process is None:
        return (
            {"detail": "Failed to run agent {}".format(agent_id)},
            400,
        )  # 400 Bad request
    app_context.agent_processes[agent_id] = agent_process
    app_context.agent_tty[agent_id] = []
    app_context.agent_error[agent_id] = []

    tty_read_thread = threading.Thread(
        target=read_tty,
        args=(app_context.agent_processes[agent_id], app_context.agent_tty[agent_id],),
    )
    tty_read_thread.start()

    error_read_thread = threading.Thread(
        target=read_error,
        args=(
            app_context.agent_processes[agent_id],
            app_context.agent_error[agent_id],
        ),
    )
    error_read_thread.start()

    return agent_id, 201  # 200 (OK)


def get_agent_status(agent_id: str):
    """Get the status of the running agent Node."""
    status_str = str(ProcessState.NOT_STARTED).replace("ProcessState.", "")
    tty_str = ""
    error_str = ""

    # agent_id will not be in lists if we haven't run it yet
    if (
        agent_id in app_context.agent_processes
        and app_context.agent_processes[agent_id] is not None
    ):
        status_str = str(
            get_process_status(app_context.agent_processes[agent_id])
        ).replace("ProcessState.", "")

    if agent_id in app_context.agent_tty:
        total_num_lines = len(app_context.agent_tty[agent_id])
        for i in range(max(0, total_num_lines - max_log_lines), total_num_lines):
            tty_str += app_context.agent_tty[agent_id][i]

    else:
        tty_str = ""

    tty_str = tty_str.replace("\n", "<br>")

    if agent_id in app_context.agent_error:
        total_num_lines = len(app_context.agent_error[agent_id])
        for i in range(max(0, total_num_lines - max_log_lines), total_num_lines):
            error_str += app_context.agent_error[agent_id][i]

    else:
        error_str = ""

    error_str = error_str.replace("\n", "<br>")

    return {"status": status_str, "tty": tty_str, "error": error_str}, 200  # (OK)


def stop_agent(agent_id: str):
    """Stop agent running."""
    # pass to private function to make it easier to mock
    return stop_agent_process(agent_id, app_context)


def create_app():
    """Run the flask server."""
    CUR_DIR = os.path.abspath(os.path.dirname(__file__))
    app = connexion.FlaskApp(__name__, specification_dir=CUR_DIR)
    global app_context  # pylint: disable=global-statement
    app_context = AppContext()

    app_context.agent_processes = {}
    app_context.agent_tty = {}
    app_context.agent_error = {}
    app_context.ui_is_starting = False
    app_context.agents_dir = os.path.abspath(os.getcwd())
    app_context.module_dir = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "../../"
    )

    app.add_api("aea_cli_rest.yaml")

    @app.route("/")
    def home():  # pylint: disable=unused-variable
        """Respond to browser URL:  localhost:5000/."""
        return flask.render_template(
            "home.html", len=len(elements), htmlElements=elements
        )

    @app.route("/static/js/home.js")
    def homejs():  # pylint: disable=unused-variable
        """Serve the home.js file (as it needs templating)."""
        return flask.render_template(
            "home.js", len=len(elements), htmlElements=elements
        )

    @app.route("/favicon.ico")
    def favicon():  # pylint: disable=unused-variable
        """Return an icon to be displayed in the browser."""
        return flask.send_from_directory(
            os.path.join(app.root_path, "static"),
            "favicon.ico",
            mimetype="image/vnd.microsoft.icon",
        )

    return app


def run(port: int, host: str = "127.0.0.1"):
    """Run the GUI."""

    app = create_app()
    try:
        app.run(host=host, port=port, debug=False)
    finally:
        terminate_processes()

    return app


def run_test():
    """Run the gui in the form where we can run tests against it."""
    app = create_app()
    return app.app.test_client()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the code-blocks in the skill.md file."""
from pathlib import Path

import mistune

from aea.skills.behaviours import OneShotBehaviour
from aea.skills.tasks import Task

from tests.conftest import ROOT_DIR
from tests.test_docs.helper import compile_and_exec


class TestSkillDocs:
    """Test the integrity of the code-blocks in skill.md"""

    @classmethod
    def setup_class(cls):
        """Test skill.md"""
        markdown_parser = mistune.create_markdown(renderer=mistune.AstRenderer())

        skill_doc_file = Path(ROOT_DIR, "docs", "skill.md")
        doc = markdown_parser(skill_doc_file.read_text())
        # get only code blocks
        cls.code_blocks = list(filter(lambda x: x["type"] == "block_code", doc))

    def test_context(self):
        """Test the code in context."""
        block = self.code_blocks[0]
        expected = "self.context.outbox.put_message(message=reply)"
        assert block["text"].strip() == expected
        assert block["info"].strip() == "python"

    # TODO add tests for new_handlers queue

    def test_hello_world_behaviour(self):
        """Test the code in the 'behaviours.py' section."""
        # here, we test the definition of a custom class
        offset = 2
        block = self.code_blocks[offset]
        text = block["text"]

        # check that the code can be executed
        code_obj = compile(text, "fakemodule", "exec")
        locals_dict = {}
        exec(code_obj, globals(), locals_dict)  # nosec

        # some consistency check on the behaviour class.
        HelloWorldBehaviour = locals_dict["HelloWorldBehaviour"]
        assert issubclass(HelloWorldBehaviour, OneShotBehaviour)

        # here, we test the code example for adding the new custom behaviour to the list
        # of new behaviours
        block = self.code_blocks[offset + 1]
        text = block["text"]
        assert (
            text.strip()
            == 'self.context.new_behaviours.put(HelloWorldBehaviour(name="hello_world", skill_context=self.context))'
        )

        block = self.code_blocks[offset + 2]
        assert (
            block["text"] == "def hello():\n"
            '    print("Hello, World!")\n'
            "\n"
            'self.context.new_behaviours.put(OneShotBehaviour(act=hello, name="hello_world", skill_context=self.context))\n'
        )

    def test_task(self):
        """Test the code blocks of the 'tasks.py' section."""
        # test code of task definition
        offset = 5
        block = self.code_blocks[offset]
        locals_dict = compile_and_exec(block["text"])

        nth_prime_number = locals_dict["nth_prime_number"]
        assert nth_prime_number(1) == 2
        assert nth_prime_number(2) == 3
        assert nth_prime_number(3) == 5
        assert nth_prime_number(4) == 7
        LongTask = locals_dict["LongTask"]
        assert issubclass(LongTask, Task)
        LongTask()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This CLI tool takes the main dependencies of the Pipfile.lock and prints it to stdout in requirements.txt format."""
import json


def parse_args():
    """Parse CLI arguments."""
    import argparse  # pylint: disable=import-outside-toplevel

    parser = argparse.ArgumentParser("parse_main_dependencies_from_lock")
    parser.add_argument(
        "pipfile_lock_path", type=argparse.FileType("r"), help="Path to Pipfile.lock."
    )
    parser.add_argument("-o", "--output", type=argparse.FileType("w"), default=None)
    return parser.parse_args()


if __name__ == "__main__":
    arguments = parse_args()

    pipfile_lock_content = json.load(arguments.pipfile_lock_path)
    requirements = sorted(
        map(
            lambda x: x[0] + x[1]["version"],
            pipfile_lock_content.get("default").items(),
        )
    )

    requirements_content = "\n".join(requirements)
    if arguments.output is None:
        print(requirements_content)
    else:
        arguments.output.write(requirements_content)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tasks for the 'ml_train' skill."""

from typing import Tuple

import numpy as np
from tensorflow import keras

from aea.skills.base import SkillContext
from aea.skills.tasks import Task


class MLTrainTask(Task):
    """ML train task."""

    def __init__(
        self,
        skill_context: SkillContext,
        train_data: Tuple[np.ndarray, np.ndarray],
        model: keras.Model,
        epochs_per_batch: int = 10,
        batch_size: int = 32,
    ):
        """Initialize the task."""
        super().__init__(logger=skill_context.logger)
        self.train_x, self.train_y = train_data

        self.model = model
        self.epochs_per_batch = epochs_per_batch
        self.batch_size = batch_size

    def setup(self) -> None:
        """Set up the task."""
        self.logger.info("ML Train task: setup method called.")

    def execute(self, *args, **kwargs) -> keras.Model:
        """Execute the task."""
        self.logger.info("Start training with {} rows".format(self.train_x.shape[0]))
        self.model.fit(self.train_x, self.train_y, epochs=self.epochs_per_batch)
        loss, acc = self.model.evaluate(self.train_x, self.train_y, verbose=2)
        self.logger.info("Loss: {}, Acc: {}".format(loss, acc))
        return self.model

    def teardown(self) -> None:
        """Teardown the task."""
        self.logger.info("ML Train task: teardown method called.")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""HTTP client connection and channel."""

import asyncio
import json
import logging
from asyncio import CancelledError
from asyncio.events import AbstractEventLoop
from asyncio.tasks import Task
from traceback import format_exc
from typing import Any, Optional, Set, Tuple, Type, Union, cast

import aiohttp
from aiohttp.client_reqrep import ClientResponse

from aea.common import Address
from aea.configurations.base import PublicId
from aea.connections.base import Connection, ConnectionStates
from aea.exceptions import enforce
from aea.mail.base import Envelope, EnvelopeContext, Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.protocols.dialogue.base import DialogueLabel as BaseDialogueLabel

from packages.fetchai.protocols.http.dialogues import HttpDialogue as BaseHttpDialogue
from packages.fetchai.protocols.http.dialogues import HttpDialogues as BaseHttpDialogues
from packages.fetchai.protocols.http.message import HttpMessage


SUCCESS = 200
NOT_FOUND = 404
REQUEST_TIMEOUT = 408
SERVER_ERROR = 500
PUBLIC_ID = PublicId.from_str("fetchai/http_client:0.10.0")

_default_logger = logging.getLogger("aea.packages.fetchai.connections.http_client")

RequestId = str


class HttpDialogue(BaseHttpDialogue):
    """The dialogue class maintains state of a dialogue and manages it."""

    def __init__(
        self,
        dialogue_label: BaseDialogueLabel,
        self_address: Address,
        role: BaseDialogue.Role,
        message_class: Type[HttpMessage] = HttpMessage,
    ) -> None:
        """
        Initialize a dialogue.

        :param dialogue_label: the identifier of the dialogue
        :param self_address: the address of the entity for whom this dialogue is maintained
        :param role: the role of the agent this dialogue is maintained for

        :return: None
        """
        BaseHttpDialogue.__init__(
            self,
            dialogue_label=dialogue_label,
            self_address=self_address,
            role=role,
            message_class=message_class,
        )
        self._envelope_context = None  # type: Optional[EnvelopeContext]

    @property
    def envelope_context(self) -> Optional[EnvelopeContext]:
        """Get envelope_context."""
        return self._envelope_context

    @envelope_context.setter
    def envelope_context(self, envelope_context: Optional[EnvelopeContext]) -> None:
        """Set envelope_context."""
        enforce(self._envelope_context is None, "envelope_context already set!")
        self._envelope_context = envelope_context


class HttpDialogues(BaseHttpDialogues):
    """The dialogues class keeps track of all http dialogues."""

    def __init__(self) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            # The client connection maintains the dialogue on behalf of the server
            return HttpDialogue.Role.SERVER

        BaseHttpDialogues.__init__(
            self,
            self_address=str(HTTPClientConnection.connection_id),
            role_from_first_message=role_from_first_message,
            dialogue_class=HttpDialogue,
        )


class HTTPClientAsyncChannel:
    """A wrapper for a HTTPClient."""

    DEFAULT_TIMEOUT = 300  # default timeout in seconds
    DEFAULT_EXCEPTION_CODE = (
        600  # custom code to indicate there was exception during request
    )

    def __init__(
        self,
        agent_address: Address,
        address: str,
        port: int,
        connection_id: PublicId,
        excluded_protocols: Optional[Set[PublicId]] = None,
        restricted_to_protocols: Optional[Set[PublicId]] = None,
    ):
        """
        Initialize an http client channel.

        :param agent_address: the address of the agent.
        :param address: server hostname / IP address
        :param port: server port number
        :param excluded_protocols: this connection cannot handle messages adhering to any of the protocols in this set
        :param restricted_to_protocols: this connection can only handle messages adhering to protocols in this set
        """
        self.agent_address = agent_address
        self.address = address
        self.port = port
        self.connection_id = connection_id
        self.restricted_to_protocols = restricted_to_protocols
        self._dialogues = HttpDialogues()

        self._in_queue = None  # type: Optional[asyncio.Queue]  # pragma: no cover
        self._loop = (
            None
        )  # type: Optional[asyncio.AbstractEventLoop]  # pragma: no cover
        self.excluded_protocols = excluded_protocols
        self.is_stopped = True
        self._tasks: Set[Task] = set()

        self.logger = _default_logger
        self.logger.info("Initialised the HTTP client channel")

    async def connect(self, loop: AbstractEventLoop) -> None:
        """
        Connect channel using loop.

        :param loop: asyncio event loop to use

        :return: None
        """
        self._loop = loop
        self._in_queue = asyncio.Queue()
        self.is_stopped = False

    def _get_message_and_dialogue(
        self, envelope: Envelope
    ) -> Tuple[HttpMessage, Optional[HttpDialogue]]:
        """
        Get a message copy and dialogue related to this message.

        :param envelope: incoming envelope

        :return: Tuple[MEssage, Optional[Dialogue]]
        """
        message = cast(HttpMessage, envelope.message)
        dialogue = cast(Optional[HttpDialogue], self._dialogues.update(message))
        if dialogue is not None:
            dialogue.envelope_context = envelope.context
        return message, dialogue

    async def _http_request_task(self, request_envelope: Envelope) -> None:
        """
        Perform http request and send back response.

        :param request_http_message: HttpMessage with http request constructed.

        :return: None
        """
        if not self._loop:  # pragma: nocover
            raise ValueError("Channel is not connected")

        request_http_message, dialogue = self._get_message_and_dialogue(
            request_envelope
        )

        if not dialogue:
            self.logger.warning(
                "Could not create dialogue for message={}".format(request_http_message)
            )
            return

        try:
            resp = await asyncio.wait_for(
                self._perform_http_request(request_http_message),
                timeout=self.DEFAULT_TIMEOUT,
            )
            envelope = self.to_envelope(
                request_http_message,
                status_code=resp.status,
                headers=resp.headers,
                status_text=resp.reason,
                bodyy=resp._body  # pylint: disable=protected-access
                if resp._body is not None  # pylint: disable=protected-access
                else b"",
                dialogue=dialogue,
            )
        except Exception:  # pragma: nocover # pylint: disable=broad-except
            envelope = self.to_envelope(
                request_http_message,
                status_code=self.DEFAULT_EXCEPTION_CODE,
                headers={},
                status_text="HTTPConnection request error.",
                bodyy=format_exc().encode("utf-8"),
                dialogue=dialogue,
            )

        if self._in_queue is not None:
            await self._in_queue.put(envelope)

    async def _perform_http_request(
        self, request_http_message: HttpMessage
    ) -> ClientResponse:
        """
        Perform http request and return response.

        :param request_http_message: HttpMessage with http request constructed.

        :return: aiohttp.ClientResponse
        """
        try:
            async with aiohttp.ClientSession() as session:
                async with session.request(
                    method=request_http_message.method,
                    url=request_http_message.url,
                    headers=request_http_message.headers,
                    data=request_http_message.bodyy,
                ) as resp:
                    await resp.read()
                return resp
        except Exception:  # pragma: nocover # pylint: disable=broad-except
            self.logger.exception(
                f"Exception raised during http call: {request_http_message.method} {request_http_message.url}"
            )
            raise

    def send(self, request_envelope: Envelope) -> None:
        """
        Send an envelope with http request data to request.

        Convert an http envelope into an http request.
        Send the http request
        Wait for and receive its response
        Translate the response into a response envelop.
        Send the response envelope to the in-queue.

        :param request_envelope: the envelope containing an http request

        :return: None
        """
        if self._loop is None or self.is_stopped:
            raise ValueError("Can not send a message! Channel is not started!")

        if request_envelope is None:
            return

        if request_envelope.protocol_id in (self.excluded_protocols or []):
            self.logger.error(
                "This envelope cannot be sent with the http client connection: protocol_id={}".format(
                    request_envelope.protocol_id
                )
            )
            raise ValueError("Cannot send message.")

        enforce(
            isinstance(request_envelope.message, HttpMessage),
            "Message not of type HttpMessage",
        )

        request_http_message = cast(HttpMessage, request_envelope.message)

        if (
            request_http_message.performative != HttpMessage.Performative.REQUEST
        ):  # pragma: nocover
            self.logger.warning(
                "The HTTPMessage performative must be a REQUEST. Envelop dropped."
            )
            return

        task = self._loop.create_task(self._http_request_task(request_envelope))
        task.add_done_callback(self._task_done_callback)
        self._tasks.add(task)

    def _task_done_callback(self, task: Task) -> None:
        """
        Handle http request task completed.

        Removes tasks from _tasks.

        :param task: Task completed.

        :return: None
        """
        self._tasks.remove(task)
        self.logger.debug(f"Task completed: {task}")

    async def get_message(self) -> Union["Envelope", None]:
        """
        Get http response from in-queue.

        :return: None or envelope with http response.
        """
        if self._in_queue is None:
            raise ValueError("Looks like channel is not connected!")

        try:
            return await self._in_queue.get()
        except CancelledError:  # pragma: nocover
            return None

    @staticmethod
    def to_envelope(
        http_request_message: HttpMessage,
        status_code: int,
        headers: dict,
        status_text: Optional[Any],
        bodyy: bytes,
        dialogue: HttpDialogue,
    ) -> Envelope:
        """
        Convert an HTTP response object (from the 'requests' library) into an Envelope containing an HttpMessage (from the 'http' Protocol).

        :param http_request_message: the message of the http request envelop
        :param status_code: the http status code, int
        :param headers: dict of http response headers
        :param status_text: the http status_text, str
        :param bodyy: bytes of http response content

        :return: Envelope with http response data.
        """
        http_message = dialogue.reply(
            performative=HttpMessage.Performative.RESPONSE,
            target_message=http_request_message,
            status_code=status_code,
            headers=json.dumps(dict(headers.items())),
            status_text=status_text,
            bodyy=bodyy,
            version="",
        )
        envelope = Envelope(
            to=http_message.to,
            sender=http_message.sender,
            protocol_id=http_message.protocol_id,
            context=dialogue.envelope_context,
            message=http_message,
        )
        return envelope

    async def _cancel_tasks(self) -> None:
        """Cancel all requests tasks pending."""
        for task in list(self._tasks):
            if task.done():  # pragma: nocover
                continue
            task.cancel()

        for task in list(self._tasks):
            try:
                await task
            except KeyboardInterrupt:  # pragma: nocover
                raise
            except BaseException:  # pragma: nocover # pylint: disable=broad-except
                pass  # nosec

    async def disconnect(self) -> None:
        """Disconnect."""
        if not self.is_stopped:
            self.logger.info("HTTP Client has shutdown on port: {}.".format(self.port))
            self.is_stopped = True

            await self._cancel_tasks()


class HTTPClientConnection(Connection):
    """Proxy to the functionality of the web client."""

    connection_id = PUBLIC_ID

    def __init__(self, **kwargs):
        """Initialize a HTTP client connection."""
        super().__init__(**kwargs)
        host = cast(str, self.configuration.config.get("host"))
        port = cast(int, self.configuration.config.get("port"))
        if host is None or port is None:  # pragma: nocover
            raise ValueError("host and port must be set!")
        self.channel = HTTPClientAsyncChannel(
            self.address,
            host,
            port,
            connection_id=self.connection_id,
            excluded_protocols=self.excluded_protocols,
        )

    async def connect(self) -> None:
        """
        Connect to a HTTP server.

        :return: None
        """
        if self.is_connected:  # pragma: nocover
            return

        with self._connect_context():
            self.channel.logger = self.logger
            await self.channel.connect(self.loop)

    async def disconnect(self) -> None:
        """
        Disconnect from a HTTP server.

        :return: None
        """
        if self.is_disconnected:
            return  # pragma: nocover
        self._state.set(ConnectionStates.disconnecting)
        await self.channel.disconnect()
        self._state.set(ConnectionStates.disconnected)

    async def send(self, envelope: "Envelope") -> None:
        """
        Send an envelope.

        :param envelope: the envelop
        :return: None
        """
        self._ensure_connected()
        self.channel.send(envelope)

    async def receive(self, *args, **kwargs) -> Optional[Union["Envelope", None]]:
        """
        Receive an envelope.

        :return: the envelope received, or None.
        """
        self._ensure_connected()
        try:
            return await self.channel.get_message()
        except Exception:  # pragma: nocover # pylint: disable=broad-except
            self.logger.exception("Exception on receive")
            return None



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains the tests for the `aea generate protocol` sub-command."""

import json
import os
import shutil
import tempfile
import unittest.mock
from pathlib import Path

import jsonschema
import yaml
from jsonschema import Draft4Validator, ValidationError

from aea.cli import cli
from aea.configurations.base import DEFAULT_PROTOCOL_CONFIG_FILE
from aea.configurations.loader import make_jsonschema_base_uri

from tests.conftest import (
    AUTHOR,
    CLI_LOG_OPTION,
    CONFIGURATION_SCHEMA_DIR,
    CUR_PATH,
    CliRunner,
    PROTOCOL_CONFIGURATION_SCHEMA,
)


class TestGenerateProtocol:
    """Test that the command 'aea generate protocol' works correctly in correct preconditions."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        shutil.copyfile(
            Path(CUR_PATH, "data", "sample_specification.yaml"),
            Path(cls.t, "sample_specification.yaml"),
        )
        cls.path_to_specification = str(Path("..", "sample_specification.yaml"))

        cls.schema = json.load(open(PROTOCOL_CONFIGURATION_SCHEMA))
        cls.resolver = jsonschema.RefResolver(
            make_jsonschema_base_uri(Path(CONFIGURATION_SCHEMA_DIR).absolute()),
            cls.schema,
        )
        cls.validator = Draft4Validator(cls.schema, resolver=cls.resolver)

        # create an agent
        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )
        assert result.exit_code == 0

        cls.create_result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        os.chdir(cls.agent_name)

        # generate protocol
        cls.result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "generate", "protocol", cls.path_to_specification],
            standalone_mode=False,
        )
        os.chdir(cls.cwd)

    def test_create_agent_exit_code_equal_to_0(self):
        """Test that the exit code is equal to 0 when creating the agent."""
        assert self.create_result.exit_code == 0

    def test_exit_code_equal_to_0(self):
        """Test that the exit code is equal to 0 when generating a protocol."""
        assert self.result.exit_code == 0, "Failed with stdout='{}'".format(
            self.result.stdout
        )

    def test_resource_folder_contains_configuration_file(self):
        """Test that the protocol folder contains a structurally valid configuration file."""
        p = Path(
            self.t,
            self.agent_name,
            "protocols",
            "t_protocol",
            DEFAULT_PROTOCOL_CONFIG_FILE,
        )
        config_file = yaml.safe_load(open(p))
        self.validator.validate(instance=config_file)

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestGenerateProtocolFailsWhenDirectoryAlreadyExists:
    """Test that the command 'aea generate protocol' fails when a directory with the same name as the name of the protocol being generated already exists."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.protocol_name = "t_protocol"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        shutil.copyfile(
            Path(CUR_PATH, "data", "sample_specification.yaml"),
            Path(cls.t, "sample_specification.yaml"),
        )
        cls.path_to_specification = str(Path("..", "sample_specification.yaml"))

        # create an agent
        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )
        assert result.exit_code == 0

        cls.create_result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        os.chdir(cls.agent_name)

        # create a dummy 'myprotocol' folder
        Path(cls.t, cls.agent_name, "protocols", cls.protocol_name).mkdir(
            exist_ok=False, parents=True
        )

        # generate protocol
        cls.result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "generate", "protocol", cls.path_to_specification],
            standalone_mode=False,
        )
        os.chdir(cls.cwd)

    def test_create_agent_exit_code_equal_to_0(self):
        """Test that the exit code is equal to 0 when creating the agent."""
        assert self.create_result.exit_code == 0

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    def test_error_message_protocol_already_existing(self):
        """Test that the log error message is fixed.

        The expected message is: 'A protocol with name '{protocol_name}' already exists. Aborting...'
        """
        s = "A directory with name '{}' already exists. Aborting...".format(
            self.protocol_name
        )
        assert self.result.exception.message == s

    def test_resource_directory_exists(self):
        """Test that the resource directory still exists.

        This means that after every failure, we make sure we restore the previous state.
        """
        assert Path(self.t, self.agent_name, "protocols", self.protocol_name).exists()

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestGenerateProtocolFailsWhenProtocolAlreadyExists:
    """Test that the command 'aea add protocol' fails when the protocol already exists."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        shutil.copyfile(
            Path(CUR_PATH, "data", "sample_specification.yaml"),
            Path(cls.t, "sample_specification.yaml"),
        )
        cls.path_to_specification = str(Path("..", "sample_specification.yaml"))

        # create an agent
        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )
        assert result.exit_code == 0

        cls.create_result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        os.chdir(cls.agent_name)

        # generate protocol first time
        cls.generate_result_1 = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "generate", "protocol", cls.path_to_specification],
            standalone_mode=False,
        )

        # generate protocol second time
        cls.generate_result_2 = cls.runner.invoke(
            cli,
            [
                *CLI_LOG_OPTION,
                "--skip-consistency-check",
                "generate",
                "protocol",
                cls.path_to_specification,
            ],
            standalone_mode=False,
        )
        os.chdir(cls.cwd)

    def test_create_agent_exit_code_equal_to_0(self):
        """Test that the exit code is equal to 0 when creating the agent."""
        assert self.create_result.exit_code == 0

    def test_generate_protocol_first_time_exit_code_equal_to_0(self):
        """Test that the exit code is equal to 0 the first time a protocol is generated."""
        assert self.generate_result_1.exit_code == 0

    def test_generate_protocol_second_time_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 the second time the protocol is generated (i.e. catchall for general errors)."""
        assert self.generate_result_2.exit_code == 1

    def test_error_message_protocol_already_existing(self):
        """Test that the log error message is fixed.

        The expected message is: 'A protocol with name '{protocol_name}' already exists. Aborting...'
        """
        s = "A protocol with name 't_protocol' already exists. Aborting..."
        assert self.generate_result_2.exception.message == s

    def test_resource_directory_exists(self):
        """Test that the resource directory still exists.

        This means that after every failure, we make sure we restore the previous state.
        """
        assert Path(self.t, self.agent_name, "protocols", "t_protocol").exists()

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestGenerateProtocolFailsWhenConfigFileIsNotCompliant:
    """Test that the command 'aea generate protocol' fails when the configuration file is not compliant with the schema."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        shutil.copyfile(
            Path(CUR_PATH, "data", "sample_specification.yaml"),
            Path(cls.t, "sample_specification.yaml"),
        )
        cls.path_to_specification = str(Path("..", "sample_specification.yaml"))

        # create an agent
        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )
        assert result.exit_code == 0

        cls.create_result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )

        # change the dumping of yaml module to raise an exception.
        cls.patch = unittest.mock.patch(
            "yaml.dump_all", side_effect=ValidationError("test error message")
        )
        cls.patch.start()

        # generate protocol
        os.chdir(cls.agent_name)
        cls.result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "generate", "protocol", cls.path_to_specification],
            standalone_mode=False,
        )
        os.chdir(cls.cwd)

    def test_create_agent_exit_code_equal_to_0(self):
        """Test that the exit code is equal to 0 when creating the agent."""
        assert self.create_result.exit_code == 0

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 when config file is non-compliant (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    def test_resource_directory_does_not_exists(self):
        """Test that the resource directory does not exist.

        This means that after every failure, we make sure we restore the previous state.
        """
        assert not Path(self.t, self.agent_name, "protocols", "t_protocol").exists()

    def test_configuration_file_not_valid(self):
        """Test that the log error message is fixed.

        The expected message is: 'Cannot find protocol: '{protocol_name}'
        """
        s = "Protocol is NOT generated. The following error happened while generating the protocol:\ntest error message"
        assert self.result.exception.message == s

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        cls.patch.stop()
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestGenerateProtocolFailsWhenExceptionOccurs:
    """Test that the command 'aea generate protocol' fails when the configuration file is not compliant with the schema."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        cls.path_to_specification = str(Path("..", "sample_specification.yaml"))

        # create an agent
        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )
        assert result.exit_code == 0

        cls.create_result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )

        # create an exception
        cls.patch = unittest.mock.patch(
            "shutil.copytree", side_effect=Exception("unknwon exception")
        )
        cls.patch.start()

        # generate protocol
        os.chdir(cls.agent_name)
        cls.result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "generate", "protocol", cls.path_to_specification],
            standalone_mode=False,
        )
        os.chdir(cls.cwd)

    def test_create_agent_exit_code_equal_to_0(self):
        """Test that the exit code is equal to 0 when creating the agent."""
        assert self.create_result.exit_code == 0

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 when an exception is thrown (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    def test_resource_directory_does_not_exists(self):
        """Test that the resource directory does not exist.

        This means that after every failure, we make sure we restore the previous state.
        """
        assert not Path(self.t, self.agent_name, "protocols", "t_protocol").exists()

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        cls.patch.stop()
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for aea.cli.utils module."""

from builtins import FileNotFoundError
from typing import cast
from unittest import TestCase, mock

import pytest
from click import BadParameter, ClickException
from jsonschema import ValidationError
from yaml import YAMLError

from aea.cli.utils.click_utils import PublicIdParameter
from aea.cli.utils.config import (
    _init_cli_config,
    get_or_create_cli_config,
    update_cli_config,
)
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import _validate_config_consistency, clean_after
from aea.cli.utils.formatting import format_items
from aea.cli.utils.generic import is_readme_present
from aea.cli.utils.package_utils import (
    find_item_in_distribution,
    find_item_locally,
    get_package_path_unified,
    get_wallet_from_context,
    is_fingerprint_correct,
    is_item_present_unified,
    is_local_item,
    try_get_balance,
    try_get_item_source_path,
    try_get_item_target_path,
    validate_author_name,
    validate_package_name,
)
from aea.configurations.base import PublicId
from aea.configurations.constants import (
    DEFAULT_CONNECTION,
    DEFAULT_PROTOCOL,
    DEFAULT_SKILL,
    SIGNING_PROTOCOL,
    STATE_UPDATE_PROTOCOL,
)
from aea.crypto.wallet import Wallet
from aea.helpers.base import cd
from aea.test_tools.test_cases import AEATestCaseEmpty

from tests.conftest import FETCHAI
from tests.test_cli.tools_for_testing import (
    ConfigLoaderMock,
    ContextMock,
    PublicIdMock,
    StopTest,
    raise_stoptest,
)


AUTHOR = "author"


class FormatItemsTestCase(TestCase):
    """Test case for format_items method."""

    def testformat_items_positive(self):
        """Test format_items positive result."""
        items = [
            {
                "public_id": "author/name:version",
                "name": "obj-name",
                "description": "Some description",
                "author": "author",
                "version": "1.0",
            }
        ]
        result = format_items(items)
        expected_result = (
            "------------------------------\n"
            "Public ID: author/name:version\n"
            "Name: obj-name\n"
            "Description: Some description\n"
            "Author: author\n"
            "Version: 1.0\n"
            "------------------------------\n"
        )
        self.assertEqual(result, expected_result)


@mock.patch("aea.cli.utils.package_utils.os.path.join", return_value="some-path")
class TryGetItemSourcePathTestCase(TestCase):
    """Test case for try_get_item_source_path method."""

    @mock.patch("aea.cli.utils.package_utils.os.path.exists", return_value=True)
    def test_get_item_source_path_positive(self, exists_mock, join_mock):
        """Test for get_item_source_path positive result."""
        result = try_get_item_source_path("cwd", AUTHOR, "skills", "skill-name")
        expected_result = "some-path"
        self.assertEqual(result, expected_result)
        join_mock.assert_called_once_with("cwd", AUTHOR, "skills", "skill-name")
        exists_mock.assert_called_once_with("some-path")

        result = try_get_item_source_path("cwd", None, "skills", "skill-name")
        self.assertEqual(result, expected_result)

    @mock.patch("aea.cli.utils.package_utils.os.path.exists", return_value=False)
    def test_get_item_source_path_not_exists(self, exists_mock, join_mock):
        """Test for get_item_source_path item already exists."""
        with self.assertRaises(ClickException):
            try_get_item_source_path("cwd", AUTHOR, "skills", "skill-name")


@mock.patch("aea.cli.utils.package_utils.os.path.join", return_value="some-path")
class TryGetItemTargetPathTestCase(TestCase):
    """Test case for try_get_item_target_path method."""

    @mock.patch("aea.cli.utils.package_utils.os.path.exists", return_value=False)
    def test_get_item_target_path_positive(self, exists_mock, join_mock):
        """Test for get_item_source_path positive result."""
        result = try_get_item_target_path("packages", AUTHOR, "skills", "skill-name")
        expected_result = "some-path"
        self.assertEqual(result, expected_result)
        join_mock.assert_called_once_with("packages", AUTHOR, "skills", "skill-name")
        exists_mock.assert_called_once_with("some-path")

    @mock.patch("aea.cli.utils.package_utils.os.path.exists", return_value=True)
    def test_get_item_target_path_already_exists(self, exists_mock, join_mock):
        """Test for get_item_target_path item already exists."""
        with self.assertRaises(ClickException):
            try_get_item_target_path("skills", AUTHOR, "skill-name", "packages_path")


class PublicIdParameterTestCase(TestCase):
    """Test case for PublicIdParameter class."""

    def test_get_metavar_positive(self):
        """Test for get_metavar positive result."""
        result = PublicIdParameter.get_metavar("obj", "param")
        expected_result = "PUBLIC_ID"
        self.assertEqual(result, expected_result)


@mock.patch("aea.cli.utils.config.os.path.dirname", return_value="dir-name")
@mock.patch("aea.cli.utils.config.os.path.exists", return_value=False)
@mock.patch("aea.cli.utils.config.os.makedirs")
@mock.patch("builtins.open")
class InitConfigFolderTestCase(TestCase):
    """Test case for _init_cli_config method."""

    def test_init_cli_config_positive(
        self, open_mock, makedirs_mock, exists_mock, dirname_mock
    ):
        """Test for _init_cli_config method positive result."""
        _init_cli_config()
        dirname_mock.assert_called_once()
        exists_mock.assert_called_once_with("dir-name")
        makedirs_mock.assert_called_once_with("dir-name")


@mock.patch("aea.cli.utils.config.get_or_create_cli_config")
@mock.patch("aea.cli.utils.generic.yaml.dump")
@mock.patch("builtins.open", mock.mock_open())
class UpdateCLIConfigTestCase(TestCase):
    """Test case for update_cli_config method."""

    def testupdate_cli_config_positive(self, dump_mock, icf_mock):
        """Test for update_cli_config method positive result."""
        update_cli_config({"some": "config"})
        icf_mock.assert_called_once()
        dump_mock.assert_called_once()


def _raise_yamlerror(*args):
    raise YAMLError()


def _raise_file_not_found_error(*args):
    raise FileNotFoundError()


@mock.patch("builtins.open", mock.mock_open())
class GetOrCreateCLIConfigTestCase(TestCase):
    """Test case for read_cli_config method."""

    @mock.patch(
        "aea.cli.utils.generic.yaml.safe_load", return_value={"correct": "output"}
    )
    def testget_or_create_cli_config_positive(self, safe_load_mock):
        """Test for get_or_create_cli_config method positive result."""
        result = get_or_create_cli_config()
        expected_result = {"correct": "output"}
        self.assertEqual(result, expected_result)
        safe_load_mock.assert_called_once()

    @mock.patch("aea.cli.utils.generic.yaml.safe_load", _raise_yamlerror)
    def testget_or_create_cli_config_bad_yaml(self):
        """Test for rget_or_create_cli_config method bad yaml behavior."""
        with self.assertRaises(ClickException):
            get_or_create_cli_config()


class CleanAfterTestCase(TestCase):
    """Test case for clean_after decorator method."""

    @mock.patch("aea.cli.utils.decorators.os.path.exists", return_value=True)
    @mock.patch("aea.cli.utils.decorators._cast_ctx", lambda x: x)
    @mock.patch("aea.cli.utils.decorators.shutil.rmtree")
    def test_clean_after_positive(self, rmtree_mock, *mocks):
        """Test clean_after decorator method for positive result."""

        @clean_after
        def func(click_context):
            ctx = cast(Context, click_context.obj)
            ctx.clean_paths.append("clean/path")
            raise ClickException("Message")

        with self.assertRaises(ClickException):
            func(ContextMock())
            rmtree_mock.assert_called_once_with("clean/path")


@mock.patch("aea.cli.utils.package_utils.click.echo", raise_stoptest)
class ValidateAuthorNameTestCase(TestCase):
    """Test case for validate_author_name method."""

    @mock.patch(
        "aea.cli.utils.package_utils.click.prompt", return_value="correct_author"
    )
    def test_validate_author_name_positive(self, prompt_mock):
        """Test validate_author_name for positive result."""
        author = "valid_author"
        result = validate_author_name(author=author)
        self.assertEqual(result, author)

        result = validate_author_name()
        self.assertEqual(result, "correct_author")
        prompt_mock.assert_called_once()

    @mock.patch(
        "aea.cli.utils.package_utils.click.prompt", return_value="inv@l1d_@uth&r"
    )
    def test_validate_author_name_negative(self, prompt_mock):
        """Test validate_author_name for negative result."""
        with self.assertRaises(StopTest):
            validate_author_name()

        prompt_mock.return_value = "skills"
        with self.assertRaises(StopTest):
            validate_author_name()


class ValidatePackageNameTestCase(TestCase):
    """Test case for validate_package_name method."""

    def test_validate_package_name_positive(self):
        """Test validate_package_name for positive result."""
        validate_package_name("correct_name")

    def test_validate_package_name_negative(self):
        """Test validate_package_name for negative result."""
        with self.assertRaises(BadParameter):
            validate_package_name("incorrect-name")


def _raise_validation_error(*args, **kwargs):
    raise ValidationError("Message.")


class FindItemLocallyTestCase(TestCase):
    """Test case for find_item_locally method."""

    @mock.patch("aea.cli.utils.package_utils.Path.exists", return_value=True)
    @mock.patch(
        "aea.cli.utils.package_utils.ConfigLoader.from_configuration_type",
        _raise_validation_error,
    )
    def test_find_item_locally_bad_config(self, *mocks):
        """Test find_item_locally for bad config result."""
        public_id = PublicIdMock.from_str("fetchai/echo:0.9.0")
        with self.assertRaises(ClickException) as cm:
            find_item_locally(ContextMock(), "skill", public_id)

        self.assertIn("configuration file not valid", cm.exception.message)

    @mock.patch("aea.cli.utils.package_utils.Path.exists", return_value=True)
    @mock.patch("aea.cli.utils.package_utils.Path.open", mock.mock_open())
    @mock.patch(
        "aea.cli.utils.package_utils.ConfigLoader.from_configuration_type",
        return_value=ConfigLoaderMock(),
    )
    def test_find_item_locally_cant_find(self, from_conftype_mock, *mocks):
        """Test find_item_locally for can't find result."""
        public_id = PublicIdMock.from_str("fetchai/echo:0.9.0")
        with self.assertRaises(ClickException) as cm:
            find_item_locally(ContextMock(), "skill", public_id)

        self.assertEqual(
            cm.exception.message, "Cannot find skill with author and version specified."
        )


class FindItemInDistributionTestCase(TestCase):
    """Test case for find_item_in_distribution method."""

    @mock.patch("aea.cli.utils.package_utils.Path.exists", return_value=True)
    @mock.patch(
        "aea.cli.utils.package_utils.ConfigLoader.from_configuration_type",
        _raise_validation_error,
    )
    def testfind_item_in_distribution_bad_config(self, *mocks):
        """Test find_item_in_distribution for bad config result."""
        public_id = PublicIdMock.from_str("fetchai/echo:0.9.0")
        with self.assertRaises(ClickException) as cm:
            find_item_in_distribution(ContextMock(), "skill", public_id)

        self.assertIn("configuration file not valid", cm.exception.message)

    @mock.patch("aea.cli.utils.package_utils.Path.exists", return_value=False)
    def testfind_item_in_distribution_not_found(self, *mocks):
        """Test find_item_in_distribution for not found result."""
        public_id = PublicIdMock.from_str("fetchai/echo:0.9.0")
        with self.assertRaises(ClickException) as cm:
            find_item_in_distribution(ContextMock(), "skill", public_id)

        self.assertIn("Cannot find skill", cm.exception.message)

    @mock.patch("aea.cli.utils.package_utils.Path.exists", return_value=True)
    @mock.patch("aea.cli.utils.package_utils.Path.open", mock.mock_open())
    @mock.patch(
        "aea.cli.utils.package_utils.ConfigLoader.from_configuration_type",
        return_value=ConfigLoaderMock(),
    )
    def testfind_item_in_distribution_cant_find(self, from_conftype_mock, *mocks):
        """Test find_item_locally for can't find result."""
        public_id = PublicIdMock.from_str("fetchai/echo:0.9.0")
        with self.assertRaises(ClickException) as cm:
            find_item_in_distribution(ContextMock(), "skill", public_id)

        self.assertEqual(
            cm.exception.message, "Cannot find skill with author and version specified."
        )


class ValidateConfigConsistencyTestCase(TestCase):
    """Test case for _validate_config_consistency method."""

    @mock.patch("aea.cli.utils.config.Path.exists", _raise_validation_error)
    def test__validate_config_consistency_cant_find(self, *mocks):
        """Test _validate_config_consistency can't find result"""
        with self.assertRaises(ValueError) as cm:
            _validate_config_consistency(ContextMock(protocols=["some"]))

        self.assertIn("Cannot find", str(cm.exception))


@mock.patch(
    "aea.cli.utils.package_utils._compute_fingerprint",
    return_value={"correct": "fingerprint"},
)
class IsFingerprintCorrectTestCase(TestCase):
    """Test case for adding skill with invalid fingerprint."""

    def test_is_fingerprint_correct_positive(self, *mocks):
        """Test is_fingerprint_correct method for positive result."""
        item_config = mock.Mock()
        item_config.fingerprint = {"correct": "fingerprint"}
        item_config.fingerprint_ignore_patterns = []
        result = is_fingerprint_correct("package_path", item_config)
        self.assertTrue(result)

    def test_is_fingerprint_correct_negative(self, *mocks):
        """Test is_fingerprint_correct method for negative result."""
        item_config = mock.Mock()
        item_config.fingerprint = {"incorrect": "fingerprint"}
        item_config.fingerprint_ignore_patterns = []
        package_path = "package_dir"
        result = is_fingerprint_correct(package_path, item_config)
        self.assertFalse(result)


@mock.patch("aea.cli.utils.package_utils.LedgerApis", mock.MagicMock())
class TryGetBalanceTestCase(TestCase):
    """Test case for try_get_balance method."""

    def test_try_get_balance_positive(self):
        """Test for try_get_balance method positive result."""
        agent_config = mock.Mock()
        agent_config.default_ledger_config = FETCHAI

        wallet_mock = mock.Mock()
        wallet_mock.addresses = {FETCHAI: "some-adress"}
        try_get_balance(agent_config, wallet_mock, FETCHAI)


@mock.patch("aea.cli.utils.generic.os.path.exists", return_value=True)
class IsReadmePresentTestCase(TestCase):
    """Test case for is_readme_present method."""

    def test_is_readme_present_positive(self, *mocks):
        """Test is_readme_present for positive result."""
        self.assertTrue(is_readme_present("readme/path"))


@mock.patch("aea.cli.utils.package_utils.get_package_path", return_value="some_path")
@mock.patch("aea.cli.utils.package_utils.is_item_present")
@pytest.mark.parametrize("vendor", [True, False])
def test_get_package_path_unified(mock_present, mock_path, vendor):
    """Test 'get_package_path_unified'."""
    contex_mock = mock.MagicMock()
    contex_mock.agent_config.author = "some_author" if vendor else "another_author"
    mock_present.return_value = vendor
    public_id_mock = mock.MagicMock(author="some_author")
    result = get_package_path_unified(
        contex_mock, "some_component_type", public_id_mock
    )
    assert result == "some_path"


@mock.patch("aea.cli.utils.package_utils.is_item_present", return_value=False)
@pytest.mark.parametrize("vendor", [True, False])
def test_is_item_present_unified(mock_, vendor):
    """Test 'is_item_present_unified'."""
    contex_mock = mock.MagicMock()
    contex_mock.agent_config.author = "some_author" if vendor else "another_author"
    public_id_mock = mock.MagicMock(author="some_author")
    result = is_item_present_unified(contex_mock, "some_component_type", public_id_mock)
    assert not result


@pytest.mark.parametrize(
    ["public_id", "expected_outcome"],
    [
        (PublicId.from_str("author/package:0.1.0"), False),
        (PublicId.from_str("author/package:latest"), False),
        (PublicId.from_str("fetchai/oef:0.1.0"), False),
        (PublicId.from_str("fetchai/oef:latest"), False),
        (DEFAULT_CONNECTION, True),
        (DEFAULT_SKILL, True),
        (DEFAULT_PROTOCOL, True),
        (SIGNING_PROTOCOL, True),
        (STATE_UPDATE_PROTOCOL, True),
        (DEFAULT_CONNECTION.to_latest(), True),
        (DEFAULT_SKILL.to_latest(), True),
        (DEFAULT_PROTOCOL.to_latest(), True),
        (SIGNING_PROTOCOL.to_latest(), True),
        (STATE_UPDATE_PROTOCOL.to_latest(), True),
    ],
)
def test_is_local_item(public_id, expected_outcome):
    """Test the 'is_local_item' CLI utility function."""
    assert is_local_item(public_id) is expected_outcome


class TestGetWalletFromtx(AEATestCaseEmpty):
    """Test get_wallet_from_context."""

    def test_get_wallet_from_ctx(self):
        """Test get_wallet_from_context."""
        ctx = mock.Mock()
        with cd(self._get_cwd()):
            assert isinstance(get_wallet_from_context(ctx), Wallet)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the resources class."""
from contextlib import suppress
from typing import List, Optional, cast

from aea.components.base import Component
from aea.configurations.base import (
    ComponentId,
    ComponentType,
    ConnectionId,
    ContractId,
    SkillId,
)
from aea.connections.base import Connection
from aea.contracts.base import Contract
from aea.protocols.base import Protocol
from aea.registries.base import (
    AgentComponentRegistry,
    ComponentRegistry,
    HandlerRegistry,
    ProtocolId,
    Registry,
)
from aea.skills.base import Behaviour, Handler, Model, Skill


class Resources:
    """This class implements the object that holds the resources of an AEA."""

    def __init__(self, agent_name: str = "standalone") -> None:
        """
        Instantiate the resources.

        :return None
        """
        self._agent_name = agent_name
        self._component_registry = AgentComponentRegistry(agent_name=agent_name)
        self._handler_registry = HandlerRegistry(agent_name=agent_name)
        self._behaviour_registry = ComponentRegistry[Behaviour](agent_name=agent_name)
        self._model_registry = ComponentRegistry[Model](agent_name=agent_name)

        self._registries = [
            self._component_registry,
            self._handler_registry,
            self._behaviour_registry,
            self._model_registry,
        ]  # type: List[Registry]

    @property
    def agent_name(self) -> str:
        """Get the agent name."""
        return self._agent_name

    @property
    def component_registry(self) -> AgentComponentRegistry:
        """Get the agent component registry."""
        return self._component_registry

    @property
    def behaviour_registry(self) -> ComponentRegistry[Behaviour]:
        """Get the behaviour registry."""
        return self._behaviour_registry

    @property
    def handler_registry(self) -> HandlerRegistry:
        """Get the handler registry."""
        return self._handler_registry

    @property
    def model_registry(self) -> ComponentRegistry[Model]:
        """Get the model registry."""
        return self._model_registry

    def add_component(self, component: Component):
        """Add a component to resources."""
        if component.component_type == ComponentType.PROTOCOL:
            self.add_protocol(cast(Protocol, component))
        elif component.component_type == ComponentType.SKILL:
            self.add_skill(cast(Skill, component))
        elif component.component_type == ComponentType.CONTRACT:
            self.add_contract(cast(Contract, component))
        elif component.component_type == ComponentType.CONNECTION:
            self.add_connection(cast(Connection, component))
        else:
            raise ValueError(
                "Component type {} not supported.".format(
                    component.component_type.value
                )
            )

    def add_protocol(self, protocol: Protocol) -> None:
        """
        Add a protocol to the set of resources.

        :param protocol: a protocol
        :return: None
        """
        self._component_registry.register(protocol.component_id, protocol)

    def get_protocol(self, protocol_id: ProtocolId) -> Optional[Protocol]:
        """
        Get protocol for given protocol id.

        :param protocol_id: the protocol id
        :return: a matching protocol, if present, else None
        """
        protocol = self._component_registry.fetch(
            ComponentId(ComponentType.PROTOCOL, protocol_id)
        )
        return cast(Protocol, protocol)

    def get_all_protocols(self) -> List[Protocol]:
        """
        Get the list of all the protocols.

        :return: the list of protocols.
        """
        protocols = self._component_registry.fetch_by_type(ComponentType.PROTOCOL)
        return cast(List[Protocol], protocols)

    def remove_protocol(self, protocol_id: ProtocolId) -> None:
        """
        Remove a protocol from the set of resources.

        :param protocol_id: the protocol id for the protocol to be removed.
        :return: None
        """
        self._component_registry.unregister(
            ComponentId(ComponentType.PROTOCOL, protocol_id)
        )

    def add_contract(self, contract: Contract) -> None:
        """
        Add a contract to the set of resources.

        :param contract: a contract
        :return: None
        """
        self._component_registry.register(contract.component_id, contract)

    def get_contract(self, contract_id: ContractId) -> Optional[Contract]:
        """
        Get contract for given contract id.

        :param contract_id: the contract id
        :return: a matching contract, if present, else None
        """
        contract = self._component_registry.fetch(
            ComponentId(ComponentType.CONTRACT, contract_id)
        )
        return cast(Contract, contract)

    def get_all_contracts(self) -> List[Contract]:
        """
        Get the list of all the contracts.

        :return: the list of contracts.
        """
        contracts = self._component_registry.fetch_by_type(ComponentType.CONTRACT)
        return cast(List[Contract], contracts)

    def remove_contract(self, contract_id: ContractId) -> None:
        """
        Remove a contract from the set of resources.

        :param contract_id: the contract id for the contract to be removed.
        :return: None
        """
        self._component_registry.unregister(
            ComponentId(ComponentType.CONTRACT, contract_id)
        )

    def add_connection(self, connection: Connection) -> None:
        """
        Add a connection to the set of resources.

        :param connection: a connection
        :return: None
        """
        self._component_registry.register(connection.component_id, connection)

    def get_connection(self, connection_id: ConnectionId) -> Optional[Connection]:
        """
        Get connection for given connection id.

        :param connection_id: the connection id
        :return: a matching connection, if present, else None
        """
        connection = self._component_registry.fetch(
            ComponentId(ComponentType.CONNECTION, connection_id)
        )
        return cast(Connection, connection)

    def get_all_connections(self) -> List[Connection]:
        """
        Get the list of all the connections.

        :return: the list of connections.
        """
        connections = self._component_registry.fetch_by_type(ComponentType.CONNECTION)
        return cast(List[Connection], connections)

    def remove_connection(self, connection_id: ConnectionId) -> None:
        """
        Remove a connection from the set of resources.

        :param connection_id: the connection id for the connection to be removed.
        :return: None
        """
        self._component_registry.unregister(
            ComponentId(ComponentType.CONNECTION, connection_id)
        )

    def add_skill(self, skill: Skill) -> None:
        """
        Add a skill to the set of resources.

        :param skill: a skill
        :return: None
        """
        self._component_registry.register(skill.component_id, skill)
        if skill.handlers is not None:
            for handler in skill.handlers.values():
                self._handler_registry.register(
                    (skill.public_id, handler.name), handler
                )
        if skill.behaviours is not None:
            for behaviour in skill.behaviours.values():
                self._behaviour_registry.register(
                    (skill.public_id, behaviour.name), behaviour
                )
        if skill.models is not None:
            for model in skill.models.values():
                self._model_registry.register((skill.public_id, model.name), model)

    def get_skill(self, skill_id: SkillId) -> Optional[Skill]:
        """
        Get the skill for a given skill id.

        :param skill_id: the skill id
        :return: a matching skill, if present, else None
        """
        skill = self._component_registry.fetch(
            ComponentId(ComponentType.SKILL, skill_id)
        )
        return cast(Skill, skill)

    def get_all_skills(self) -> List[Skill]:
        """
        Get the list of all the skills.

        :return: the list of skills.
        """
        skills = self._component_registry.fetch_by_type(ComponentType.SKILL)
        return cast(List[Skill], skills)

    def remove_skill(self, skill_id: SkillId) -> None:
        """
        Remove a skill from the set of resources.

        :param skill_id: the skill id for the skill to be removed.
        :return: None
        """
        self._component_registry.unregister(ComponentId(ComponentType.SKILL, skill_id))
        with suppress(ValueError):
            self._handler_registry.unregister_by_skill(skill_id)
        with suppress(ValueError):
            self._behaviour_registry.unregister_by_skill(skill_id)
        with suppress(ValueError):
            self._model_registry.unregister_by_skill(skill_id)

    def get_handler(
        self, protocol_id: ProtocolId, skill_id: SkillId
    ) -> Optional[Handler]:
        """
        Get a specific handler.

        :param protocol_id: the protocol id the handler is handling
        :param skill_id: the skill id of the handler's skill
        :return: the handler
        """
        handler = self._handler_registry.fetch_by_protocol_and_skill(
            protocol_id, skill_id
        )
        return handler

    def get_handlers(self, protocol_id: ProtocolId) -> List[Handler]:
        """
        Get all handlers for a given protocol.

        :param protocol_id: the protocol id the handler is handling
        :return: the list of handlers matching the protocol
        """
        handlers = self._handler_registry.fetch_by_protocol(protocol_id)
        return handlers

    def get_all_handlers(self) -> List[Handler]:
        """
        Get all handlers from all skills.

        :return: the list of handlers
        """
        handlers = self._handler_registry.fetch_all()
        return handlers

    def get_behaviour(
        self, skill_id: SkillId, behaviour_name: str
    ) -> Optional[Behaviour]:
        """
        Get a specific behaviours for a given skill.

        :param skill_id: the skill id
        :param behaviour_name: the behaviour name
        :return: the behaviour, if it is present, else None
        """
        behaviour = self._behaviour_registry.fetch((skill_id, behaviour_name))
        return behaviour

    def get_behaviours(self, skill_id: SkillId) -> List[Behaviour]:
        """
        Get all behaviours for a given skill.

        :param skill_id: the skill id
        :return: the list of behaviours of the skill
        """
        behaviours = self._behaviour_registry.fetch_by_skill(
            skill_id
        )  # type: List[Behaviour]
        return behaviours

    def get_all_behaviours(self) -> List[Behaviour]:
        """
        Get all behaviours from all skills.

        :return: the list of all behaviours
        """
        behaviours = self._behaviour_registry.fetch_all()
        return behaviours

    def setup(self) -> None:
        """
        Set up the resources.

        Calls setup on all resources.

        :return: None
        """
        for r in self._registries:
            r.setup()

    def teardown(self) -> None:
        """
        Teardown the resources.

        Calls teardown on all resources.

        :return: None
        """
        for r in self._registries:
            r.teardown()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module validates a protocol specification."""

import re
from typing import Dict, List, Optional, Set, Tuple, cast

from aea.configurations.base import ProtocolSpecification
from aea.protocols.generator.common import (
    SPECIFICATION_COMPOSITIONAL_TYPES,
    SPECIFICATION_PRIMITIVE_TYPES,
    _get_sub_types_of_compositional_types,
    _has_matched_brackets,
)


# The following names are reserved for standard message fields and cannot be
# used as user defined names for performative or contents
RESERVED_NAMES = {"body", "message_id", "dialogue_reference", "target", "performative"}

# Regular expression patterns for various fields in protocol specifications
PERFORMATIVE_REGEX_PATTERN = "^[a-zA-Z0-9]+$|^[a-zA-Z0-9]+(_?[a-zA-Z0-9]+)+$"
CONTENT_NAME_REGEX_PATTERN = "^[a-zA-Z0-9]+$|^[a-zA-Z0-9]+(_?[a-zA-Z0-9]+)+$"

CT_CONTENT_TYPE_REGEX_PATTERN = "^ct:([A-Z]+[a-z]*)+$"  # or maybe "ct:(?:[A-Z][a-z]+)+" or # "^ct:[A-Z][a-zA-Z0-9]*$"

ROLE_REGEX_PATTERN = "^[a-zA-Z0-9]+$|^[a-zA-Z0-9]+(_?[a-zA-Z0-9]+)+$"
END_STATE_REGEX_PATTERN = "^[a-zA-Z0-9]+$|^[a-zA-Z0-9]+(_?[a-zA-Z0-9]+)+$"


def _is_reserved_name(content_name: str) -> bool:
    """
    Evaluate whether a content name is a reserved name or not.

    :param content_name: a content name
    :return: Boolean result
    """
    return content_name in RESERVED_NAMES


def _is_valid_regex(regex_pattern: str, text: str) -> bool:
    """
    Evaluate whether a 'text' matches a regular expression pattern.

    :param regex_pattern: the regular expression pattern
    :param text: the text on which to match regular expression

    :return: Boolean result
    """
    match = re.match(regex_pattern, text)
    return match is not None


def _has_brackets(content_type: str) -> bool:
    """
    Evaluate whether a compositional content type in a protocol specification is valid has corresponding brackets.

    :param content_type: an 'set' content type.
    :return: Boolean result
    """
    for compositional_type in SPECIFICATION_COMPOSITIONAL_TYPES:
        if content_type.startswith(compositional_type):
            content_type = content_type[len(compositional_type) :]
            if len(content_type) < 2:
                return False
            return content_type[0] == "[" and content_type[len(content_type) - 1] == "]"
    raise SyntaxError("Content type must be a compositional type!")


def _is_valid_ct(content_type: str) -> bool:
    """
    Evaluate whether the format of a 'ct' content type in a protocol specification is valid.

    :param content_type: a 'ct' content type.
    :return: Boolean result
    """
    content_type = content_type.strip()
    return _is_valid_regex(CT_CONTENT_TYPE_REGEX_PATTERN, content_type)


def _is_valid_pt(content_type: str) -> bool:
    """
    Evaluate whether the format of a 'pt' content type in a protocol specification is valid.

    :param content_type: a 'pt' content type.
    :return: Boolean result
    """
    content_type = content_type.strip()
    return content_type in SPECIFICATION_PRIMITIVE_TYPES


def _is_valid_set(content_type: str) -> bool:
    """
    Evaluate whether the format of a 'set' content type in a protocol specification is valid.

    :param content_type: a 'set' content type.
    :return: Boolean result
    """
    content_type = content_type.strip()

    if not content_type.startswith("pt:set"):
        return False

    if not _has_matched_brackets(content_type):
        return False

    if not _has_brackets(content_type):
        return False

    sub_types = _get_sub_types_of_compositional_types(content_type)
    if len(sub_types) != 1:
        return False

    sub_type = sub_types[0]
    return _is_valid_pt(sub_type)


def _is_valid_list(content_type: str) -> bool:
    """
    Evaluate whether the format of a 'list' content type in a protocol specification is valid.

    :param content_type: a 'list' content type.
    :return: Boolean result
    """
    content_type = content_type.strip()

    if not content_type.startswith("pt:list"):
        return False

    if not _has_matched_brackets(content_type):
        return False

    if not _has_brackets(content_type):
        return False

    sub_types = _get_sub_types_of_compositional_types(content_type)
    if len(sub_types) != 1:
        return False

    sub_type = sub_types[0]
    return _is_valid_pt(sub_type)


def _is_valid_dict(content_type: str) -> bool:
    """
    Evaluate whether the format of a 'dict' content type in a protocol specification is valid.

    :param content_type: a 'dict' content type.
    :return: Boolean result
    """
    content_type = content_type.strip()

    if not content_type.startswith("pt:dict"):
        return False

    if not _has_matched_brackets(content_type):
        return False

    if not _has_brackets(content_type):
        return False

    sub_types = _get_sub_types_of_compositional_types(content_type)
    if len(sub_types) != 2:
        return False

    sub_type_1 = sub_types[0]
    sub_type_2 = sub_types[1]
    return _is_valid_pt(sub_type_1) and _is_valid_pt(sub_type_2)


def _is_valid_union(content_type: str) -> bool:
    """
    Evaluate whether the format of a 'union' content type in a protocol specification is valid.

    :param content_type: an 'union' content type.
    :return: Boolean result
    """
    content_type = content_type.strip()

    if not content_type.startswith("pt:union"):
        return False

    if not _has_matched_brackets(content_type):
        return False

    if not _has_brackets(content_type):
        return False

    sub_types = _get_sub_types_of_compositional_types(content_type)
    # check there are at least two subtypes in the union
    if len(sub_types) < 2:
        return False

    # check there are no duplicate subtypes in the union
    sub_types_set = set(sub_types)
    if len(sub_types) != len(sub_types_set):
        return False

    for sub_type in sub_types:
        if not (
            _is_valid_ct(sub_type)
            or _is_valid_pt(sub_type)
            or _is_valid_set(sub_type)
            or _is_valid_list(sub_type)
            or _is_valid_dict(sub_type)
        ):
            return False

    return True


def _is_valid_optional(content_type: str) -> bool:
    """
    Evaluate whether the format of an 'optional' content type in a protocol specification is valid.

    :param content_type: an 'optional' content type.
    :return: Boolean result
    """
    content_type = content_type.strip()

    if not content_type.startswith("pt:optional"):
        return False

    if not _has_matched_brackets(content_type):
        return False

    if not _has_brackets(content_type):
        return False

    sub_types = _get_sub_types_of_compositional_types(content_type)
    if len(sub_types) != 1:
        return False

    sub_type = sub_types[0]
    return (
        _is_valid_ct(sub_type)
        or _is_valid_pt(sub_type)
        or _is_valid_set(sub_type)
        or _is_valid_list(sub_type)
        or _is_valid_dict(sub_type)
        or _is_valid_union(sub_type)
    )


def _is_valid_content_type_format(content_type: str) -> bool:
    """
    Evaluate whether the format of a content type in a protocol specification is valid.

    :param content_type: a content type.
    :return: Boolean result
    """
    return (
        _is_valid_ct(content_type)
        or _is_valid_pt(content_type)
        or _is_valid_set(content_type)
        or _is_valid_list(content_type)
        or _is_valid_dict(content_type)
        or _is_valid_union(content_type)
        or _is_valid_optional(content_type)
    )


def _validate_performatives(performative: str) -> Tuple[bool, str]:
    """
    Evaluate whether a performative in a protocol specification is valid.

    :param performative: a performative.
    :return: Boolean result, and associated message.
    """
    if not _is_valid_regex(PERFORMATIVE_REGEX_PATTERN, performative):
        return (
            False,
            "Invalid name for performative '{}'. Performative names must match the following regular expression: {} ".format(
                performative, PERFORMATIVE_REGEX_PATTERN
            ),
        )

    if _is_reserved_name(performative):
        return (
            False,
            "Invalid name for performative '{}'. This name is reserved.".format(
                performative,
            ),
        )

    return True, "Performative '{}' is valid.".format(performative)


def _validate_content_name(content_name: str, performative: str) -> Tuple[bool, str]:
    """
    Evaluate whether the name of a content in a protocol specification is valid.

    :param content_name: a content name.
    :param performative: the performative the content belongs to.

    :return: Boolean result, and associated message.
    """
    if not _is_valid_regex(CONTENT_NAME_REGEX_PATTERN, content_name):
        return (
            False,
            "Invalid name for content '{}' of performative '{}'. Content names must match the following regular expression: {} ".format(
                content_name, performative, CONTENT_NAME_REGEX_PATTERN
            ),
        )

    if _is_reserved_name(content_name):
        return (
            False,
            "Invalid name for content '{}' of performative '{}'. This name is reserved.".format(
                content_name, performative,
            ),
        )

    return (
        True,
        "Content name '{}' of performative '{}' is valid.".format(
            content_name, performative
        ),
    )


def _validate_content_type(
    content_type: str, content_name: str, performative: str
) -> Tuple[bool, str]:
    """
    Evaluate whether the type of a content in a protocol specification is valid.

    :param content_type: a content type.
    :param content_name: a content name.
    :param performative: the performative the content belongs to.

    :return: Boolean result, and associated message.
    """
    if not _is_valid_content_type_format(content_type):
        return (
            False,
            "Invalid type for content '{}' of performative '{}'. See documentation for the correct format of specification types.".format(
                content_name, performative,
            ),
        )

    return (
        True,
        "Type of content '{}' of performative '{}' is valid.".format(
            content_name, performative
        ),
    )


def _validate_speech_acts_section(
    protocol_specification: ProtocolSpecification,
) -> Tuple[bool, str, Optional[Set[str]], Optional[Set[str]]]:
    """
    Evaluate whether speech-acts of a protocol specification is valid.

    :param protocol_specification: a protocol specification.
    :return: Boolean result, associated message, set of all performatives (auxiliary), set of all custom types (auxiliary).
    """
    custom_types_set = set()
    performatives_set = set()

    if len(protocol_specification.speech_acts.read_all()) == 0:
        return (
            False,
            "Speech-acts cannot be empty!",
            None,
            None,
        )

    for (
        performative,
        speech_act_content_config,
    ) in protocol_specification.speech_acts.read_all():

        # Validate performative name
        (
            result_performative_validation,
            msg_performative_validation,
        ) = _validate_performatives(performative)
        if not result_performative_validation:
            return (
                result_performative_validation,
                msg_performative_validation,
                None,
                None,
            )

        performatives_set.add(performative)

        for content_name, content_type in speech_act_content_config.args.items():

            # Validate content name
            (
                result_content_name_validation,
                msg_content_name_validation,
            ) = _validate_content_name(content_name, performative)
            if not result_content_name_validation:
                return (
                    result_content_name_validation,
                    msg_content_name_validation,
                    None,
                    None,
                )

            # Validate content type
            (
                result_content_type_validation,
                msg_content_type_validation,
            ) = _validate_content_type(content_type, content_name, performative)
            if not result_content_type_validation:
                return (
                    result_content_type_validation,
                    msg_content_type_validation,
                    None,
                    None,
                )

            if _is_valid_ct(content_type):
                custom_types_set.add(content_type.strip())

    return True, "Speech-acts are valid.", performatives_set, custom_types_set


def _validate_protocol_buffer_schema_code_snippets(
    protocol_specification: ProtocolSpecification, custom_types_set: Set[str]
) -> Tuple[bool, str]:
    """
    Evaluate whether the protobuf code snippet section of a protocol specification is valid.

    :param protocol_specification: a protocol specification.
    :param custom_types_set: set of all custom types in the dialogue.

    :return: Boolean result, and associated message.
    """
    if (
        protocol_specification.protobuf_snippets is not None
        and protocol_specification.protobuf_snippets != ""
    ):
        for custom_type in protocol_specification.protobuf_snippets.keys():
            if custom_type not in custom_types_set:
                return (
                    False,
                    "Extra protobuf code snippet provided. Type '{}' is not used anywhere in your protocol definition.".format(
                        custom_type,
                    ),
                )
            custom_types_set.remove(custom_type)

        if len(custom_types_set) != 0:
            return (
                False,
                "No protobuf code snippet is provided for the following custom types: {}".format(
                    custom_types_set,
                ),
            )

    return True, "Protobuf code snippet section is valid."


def _validate_initiation(
    initiation: List[str], performatives_set: Set[str]
) -> Tuple[bool, str]:
    """
    Evaluate whether the initiation field in a protocol specification is valid.

    :param initiation: List of initial messages of a dialogue.
    :param performatives_set: set of all performatives in the dialogue.

    :return: Boolean result, and associated message.
    """
    if len(initiation) == 0 or initiation is None:
        return (
            False,
            "At least one initial performative for this dialogue must be specified.",
        )

    for performative in initiation:
        if performative not in performatives_set:
            return (
                False,
                "Performative '{}' specified in \"initiation\" is not defined in the protocol's speech-acts.".format(
                    performative,
                ),
            )

    return True, "Initial messages are valid."


def _validate_reply(
    reply: Dict[str, List[str]], performatives_set: Set[str]
) -> Tuple[bool, str]:
    """
    Evaluate whether the reply structure in a protocol specification is valid.

    :param reply: Reply structure of a dialogue.
    :param performatives_set: set of all performatives in the dialogue.

    :return: Boolean result, and associated message.
    """
    performatives_set_2 = performatives_set.copy()

    for performative in reply.keys():
        if performative not in performatives_set_2:
            return (
                False,
                "Performative '{}' specified in \"reply\" is not defined in the protocol's speech-acts.".format(
                    performative,
                ),
            )
        performatives_set_2.remove(performative)

    if len(performatives_set_2) != 0:
        return (
            False,
            "No reply is provided for the following performatives: {}".format(
                performatives_set_2,
            ),
        )

    return True, "Reply structure is valid."


def _validate_termination(
    termination: List[str], performatives_set: Set[str]
) -> Tuple[bool, str]:
    """
    Evaluate whether termination field in a protocol specification is valid.

    :param termination: List of terminal messages of a dialogue.
    :param performatives_set: set of all performatives in the dialogue.

    :return: Boolean result, and associated message.
    """
    if len(termination) == 0 or termination is None:
        return (
            False,
            "At least one terminal performative for this dialogue must be specified.",
        )

    for performative in termination:
        if performative not in performatives_set:
            return (
                False,
                "Performative '{}' specified in \"termination\" is not defined in the protocol's speech-acts.".format(
                    performative,
                ),
            )

    return True, "Terminal messages are valid."


def _validate_roles(roles: Set[str]) -> Tuple[bool, str]:
    """
    Evaluate whether roles field in a protocol specification is valid.

    :param roles: Set of roles of a dialogue.
    :return: Boolean result, and associated message.
    """
    if not 1 <= len(roles) <= 2:
        return (
            False,
            "There must be either 1 or 2 roles defined in this dialogue. Found {}".format(
                len(roles)
            ),
        )

    for role in roles:
        if not _is_valid_regex(ROLE_REGEX_PATTERN, role):
            return (
                False,
                "Invalid name for role '{}'. Role names must match the following regular expression: {} ".format(
                    role, ROLE_REGEX_PATTERN
                ),
            )

    return True, "Dialogue roles are valid."


def _validate_end_states(end_states: List[str]) -> Tuple[bool, str]:
    """
    Evaluate whether end_states field in a protocol specification is valid.

    :param end_states: List of end states of a dialogue.
    :return: Boolean result, and associated message.
    """
    for end_state in end_states:
        if not _is_valid_regex(END_STATE_REGEX_PATTERN, end_state):
            return (
                False,
                "Invalid name for end_state '{}'. End_state names must match the following regular expression: {} ".format(
                    end_state, END_STATE_REGEX_PATTERN
                ),
            )

    return True, "Dialogue end_states are valid."


def _validate_dialogue_section(
    protocol_specification: ProtocolSpecification, performatives_set: Set[str]
) -> Tuple[bool, str]:
    """
    Evaluate whether the dialogue section of a protocol specification is valid.

    :param protocol_specification: a protocol specification.
    :param performatives_set: set of all performatives in the dialogue.

    :return: Boolean result, and associated message.
    """
    if (
        protocol_specification.dialogue_config != {}
        and protocol_specification.dialogue_config is not None
    ):
        # Validate initiation
        result_initiation_validation, msg_initiation_validation = _validate_initiation(
            cast(List[str], protocol_specification.dialogue_config["initiation"]),
            performatives_set,
        )
        if not result_initiation_validation:
            return result_initiation_validation, msg_initiation_validation

        # Validate reply
        result_reply_validation, msg_reply_validation = _validate_reply(
            cast(Dict[str, List[str]], protocol_specification.dialogue_config["reply"]),
            performatives_set,
        )
        if not result_reply_validation:
            return result_reply_validation, msg_reply_validation

        # Validate termination
        (
            result_termination_validation,
            msg_termination_validation,
        ) = _validate_termination(
            cast(List[str], protocol_specification.dialogue_config["termination"]),
            performatives_set,
        )
        if not result_termination_validation:
            return result_termination_validation, msg_termination_validation

        # Validate roles
        result_roles_validation, msg_roles_validation = _validate_roles(
            cast(Set[str], protocol_specification.dialogue_config["roles"])
        )
        if not result_roles_validation:
            return result_roles_validation, msg_roles_validation

        # Validate end_state
        result_end_states_validation, msg_end_states_validation = _validate_end_states(
            cast(List[str], protocol_specification.dialogue_config["end_states"])
        )
        if not result_end_states_validation:
            return result_end_states_validation, msg_end_states_validation

    return True, "Dialogue section of the protocol specification is valid."


def validate(protocol_specification: ProtocolSpecification) -> Tuple[bool, str]:
    """
    Evaluate whether a protocol specification is valid.

    :param protocol_specification: a protocol specification.
    :return: Boolean result, and associated message.
    """
    # Validate speech-acts section
    (
        result_speech_acts_validation,
        msg_speech_acts_validation,
        performatives_set,
        custom_types_set,
    ) = _validate_speech_acts_section(protocol_specification)
    if not result_speech_acts_validation:
        return result_speech_acts_validation, msg_speech_acts_validation

    # Validate protocol buffer schema code snippets
    result_protobuf_validation, msg_protobuf_validation = _validate_protocol_buffer_schema_code_snippets(protocol_specification, custom_types_set)  # type: ignore
    if not result_protobuf_validation:
        return result_protobuf_validation, msg_protobuf_validation

    # Validate dialogue section
    result_dialogue_validation, msg_dialogue_validation = _validate_dialogue_section(protocol_specification, performatives_set)  # type: ignore
    if not result_dialogue_validation:
        return result_dialogue_validation, msg_dialogue_validation

    return True, "Protocol specification is valid."
